# Module with the central object that processes the data provided by the user and provides the
# basic functions to create ant-based algorithms.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
import joblib
import sys
import warnings
import random
from .ant import Ant, randomInit
from .optim import ObjectiveFunction
from .pheromone import updateAS, updateMMAS, updateACS, updateUndLocalPher, updateDirLocalPher
from .c_policy import stochasticAS
from .c_ntools import getValidPaths
from .c_utils import rouletteWheel
from .base import ScoreScaler

warnings.filterwarnings('error')

# Implemented pheromone update strategies
PHEROMONE_UPDATE_STRATEGIES = ['as', 'mmas', 'acs']


class ACO(object):
    """
    Basic object defining the parameters used by the ant-based optimisation algorithm. This object
    can be passed to some function from antco.algorithm module to execute the optimisation.

    Parameters
    ----------
    n_ants: int
        Number of ants used by the algorithm.

    graph: np.ndarray (nodes, nodes)
        Binary adjacency matrix that defines the structure of the graph to to be traversed.

    objective: antco.optim.ObjectiveFuntion subclass
        Instance of a subclass belonging to antco.optim.ObjectiveFunction that implements the
        evaluate() method in such a way that given an ant returns a scalar value.

    iterations: int
        Maximum number of iterations that the algorithm will carry out if the tolerance criterion
        has not been met before.

    heuristic: np.ndarray (nodes, nodes), dtype np.float64, default=np.ones(shape=(nodes, nodes))
        Heuristic information matrix. If no heuristic information matrix is provided, one with 
        all values equal to 1 will be created.

    **kwargs:
        evaporation: float, default=0.2
            Parameter that reference the evaporation rate of pheromones deposited on the edges of
            the graph.

        alpha: float, default=1.0
            The alpha parameter refers to the influence of pheromones when the ant makes a decision
            on the path through the walk being constructed.

        beta: float, default=1.0
            Analogous to the alpha parameter, the beta parameter refers to the importance given
            to the heuristic information received in the heuristic parameter.

        graph_type: str, default='directed'
            Parameter that indicates the representation of the graph to be explored in order to
            apply optimized algorithms for directed or undirected graphs. For undirected graphs
            use: 'undirected' or 'u'; for directed graphs use: 'directed' or 'd'.

        path_limits: tuple, default=(0, np.inf)
            Parameter that establishes the limits on the paths that ants can create in the graph.

        pheromone_init: float, default=1.0
            Pheromone initial value.

        fixed_position: bool, default=False
            Parameter indicating whether to place ants in a fixed position in the graph with no
            repeated initials positions in all the colony.

        pheromone_update: dict or str, default='AS'
            Strategy for updating pheromone values. The strategy can be provided as a string or as
            a dictionary. If it is specified as a string, no additional arguments can be included
            and they will be selected by default. In case it is provided as a dictionary additional
            parameters can be specified, in this situation the parameters can be provided as a
            dictionary where the key "strategy" will indicate the strategy to be used and the rest
            of the key-values the parameters associated to the selected strategy. The currently
            implemented strategies include:

                Ant System:

                    >>> colony = ACO(... pheromone_update='AS')

                    >>> colony = ACO(... pheromone_update={'strategy': 'AS'})

                    >>> colony = ACO(... pheromone_update={'strategy': 'AS', 'bag_size': 10,
                    >>>                                    'weight': 0.2})


                MAX-MIN Ant System:

                    >>> colony = ACO(... pheromone_update='MMAS')

                    >>> colony = ACO(... pheromone_update={'strategy': 'MMAS', 'limits': (0, 2)})

                    >>> colony = ACO(... pheromone_update={'strategy': 'MMAS', 'limits': (0, 2),
                    >>>                                    'bag_size': 10, 'weight': 0.2})

                    In this strategy, if no limits are provided these will be set to 0 and inf by
                    default.

                For more information on pheromone update methods use help(antco.undASupdate) or
                help(antco.undASupdate) for Ant System and help(antco.dirMMASupdate) or
                help(antco.undMMASupdate) for MAX-MIN Ant System.

        tol: int, default=np.inf
            This parameter corresponds to the maximum number of interactions without any improvement
            on the maximum value found before stopping the algorithm early. By default disabled.

        n_jobs: int, default=1
            Number of processes to run in parallel.

        precompute_heuristic: bool, default=True
            Indicates whether to exponentialise the values of the heuristic matrix to the beta
            parameter. If the value is set to False, the values will be exponentiated every time
            an ant performs a route (computationally more inefficient but necessary when the values
            of the heuristic matrix are not static). On the other hand, if the value is set to True,
            the values will be exponentiated only once at the beginning of the execution of the
            algorithm (computationally more efficient).

        Q: float, default None
            Parameter that determines the probability of selecting the next move deterministically 
            by selecting the move to the node that has the highest probability. By default this 
            parameter will not be considered.

        R: float, default None
            Parameter that determines the probability of selecting the next move randomly without 
            taking into account the computation of the pheromone matrix and heuristic information.
            By default  this parameter will not be considered.

        scaleScores: antco.base.ScoreScaler subclass, default None
            Subclass of antco.base.Scaler that allows scaling the scores to a certain range of
            values.

        seed: int, default=None
            Random seed.
    """

    def __init__(self, *, n_ants: int, graph: np.ndarray, objective: ObjectiveFunction, 
                 iterations: int, heuristic: np.ndarray = None, **kwargs):
        
        if heuristic is None:  # Heuristic information will have no influence on optimisation
            heuristic = np.ones(shape=graph.shape, dtype=np.int8)

        ACO.checkParameters({'n_ants': n_ants, 'graph': graph, 'heuristic': heuristic,
                             'objectiveFunction': objective, 'iterations': iterations}, kwargs)

        self._graph = np.array(graph, dtype=np.int8)
        self._H = heuristic.astype(np.float64)
        self._n_nodes = self._graph.shape[0]
        self._iterations = iterations
        self._objectiveFunction = objective
        self._updatePheromones = ACO._selectPheromoneUpdate(kwargs)
        self._pheromone_update_kw = ACO._selectPheromoneUpdateKW(kwargs)
        self._pheromone_update_kw['rho'] = kwargs.get('evaporation', 0.5)
        self._alpha = kwargs.get('alpha', 1.0)
        self._beta = kwargs.get('beta', 1.0)
        self._path_limits = kwargs.get('path_limits', (0, sys.maxsize))
        self._precompute_heuristic = kwargs.get('precompute_heuristic', True)
        self._P = np.zeros(shape=self._H.shape, dtype=np.float64)[:]
        self._P[:] = kwargs.get('pheromone_init', 1.0)
        self._pher_init = kwargs.get('pheromone_init', 1.0)
        np.fill_diagonal(self._P, 0)
        self._n_jobs = kwargs.get('n_jobs', 1)
        self._seed = kwargs.get('seed', None)
        self._tol = kwargs.get('tol', np.inf)
        self._fixed = kwargs.get('fixed_position', False)
        self._accessory_node = None
        self._Q = kwargs.get('Q', None)
        self._R = kwargs.get('R', None)
        self._scaleScores = kwargs.get('scaleScores', None)

        # Initialize ants
        self._ants = [
            Ant(l_min=self._path_limits[0], l_max=self._path_limits[1],
                graph_type=kwargs.get('graph_type', 'd')) for _ in range(n_ants)]

    def __repr__(self):
        return f'ACO(iterations={self._iterations}, pheromoneUpd={self._updatePheromones}, ' \
               f'alpha={self._alpha}, beta={self._beta}, path_limits={self._path_limits}, ' \
               f'tol={self._tol}, Q={self._Q}, R={self._R}, accessory_node={self._accessory_node}, ' \
               f'scaleScores={self._scaleScores}, n_jobs={self._n_jobs})'

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def checkParameters(args: dict, optional_args: dict):
        """ Function that checks the validity of the received arguments. """
        def _checkType(name: str, value: object, required: type):
            assert isinstance(value, required), 'Parameter: %s must be %s.' % (name, str(required))

        def _intType(name: str, value: int, min_val: int = None, max_val: int = None):
            _checkType(name, value, int)
            if min_val is not None:
                assert value > min_val, 'Parameter: %s must be grater than %d' % (name, min_val)
            if max_val is not None:
                assert value < max_val, 'Parameter: %s must be less than %d' % (name, max_val)

        def _floatType(name: str, value: float, min_val: float = None, max_val: float = None):
            _checkType(name, value, float)
            if min_val is not None:
                assert value > min_val, 'Parameter: %s must be grater than %.5f' % (name, min_val)
            if max_val is not None:
                assert value < max_val, 'Parameter: %s must be less than %.5f' % (name, max_val)

        # Check types
        _checkType('graph', args['graph'], np.ndarray)
        _checkType('heuristic', args['heuristic'], np.ndarray)
        _checkType('objectiveFunction', args['objectiveFunction'], ObjectiveFunction)
        _intType('n_ants', args['n_ants'], min_val=0)
        _intType('iterations', args['iterations'], min_val=0)

        # Check graph and heuristic dimensions (n_nodes, n_nodes)
        assert len(args['graph'].shape) == 2, \
            'Parameter graph must be a square matrix (n_nodes, n_nodes). Provided shape: {}'.\
                format(args['graph'].shape)
        assert len(args['heuristic'].shape) == 2, \
            'Parameter heuristic must be a square matrix (n_nodes, n_nodes). Provided shape: {}'.\
                format(args['heuristic'].shape)
        assert args['graph'].shape[0] == args['graph'].shape[1], \
            'Parameter graph must be a square matrix (n_nodes, n_nodes). Provided shape: {}'.\
                format(args['graph'].shape)
        assert args['heuristic'].shape[0] == args['heuristic'].shape[1], \
            'Parameter heuristic must be a square matrix (n_nodes, n_nodes). Provided shape: {}'.\
                format(args['heuristic'].shape)
        assert args['graph'].shape[0] == args['heuristic'].shape[0], \
            'The dimensions of the graph and heuristic parameters must match. Provided shapes: ' \
            'graph: %d; heuristic: %d' % (args['graph'].shape[0], args['heuristic'].shape[0])

        # Check optional arguments
        if optional_args.get('evaporation', None) is not None:
            _floatType('evaporation', optional_args['evaporation'], min_val=0.0)
        if optional_args.get('alpha', None) is not None:
            _floatType('alpha', optional_args['alpha'], min_val=0.0)
        if optional_args.get('beta', None) is not None:
            _floatType('beta', optional_args['beta'], min_val=0.0)
        if optional_args.get('fixed_position', None) is not None:
            _checkType('fixed_position', optional_args['fixed_position'], bool)
        if optional_args.get('Q', None) is not None:
            _floatType('Q', optional_args['Q'], min_val=0.0, max_val=1.0)
        if optional_args.get('R', None) is not None:
            _floatType('R', optional_args['R'], min_val=0.0, max_val=1.0)
        if optional_args.get('graph_type', None) is not None:
            assert (optional_args['graph_type'] == 'd') or (optional_args['graph_type'] == 'u') or \
                   (optional_args['graph_type'] == 'directed') or (optional_args['graph_type'] == 'undirected'), \
                'The accepted values for the graph_type parameter are: "directed" or "d" for directed ' \
                'graphs and "undirected" or "u" for undirected graphs.'
        if optional_args.get('path_limits', None) is not None:
            assert isinstance(optional_args['path_limits'], list) or \
                   isinstance(optional_args['path_limits'], tuple), 'Parameter: path_limits must be a tuple or a list'
            assert len(optional_args['path_limits']) == 2, \
                'The limits of the length of the path travelled by the ants must be defined between two integers.'
        if optional_args.get('pheromone_init', None) is not None:
            _floatType('pheromone_init', optional_args['pheromone_init'])
        if optional_args.get('tol', None) is not None:
            _intType('tol', optional_args['tol'], min_val=5)
        if optional_args.get('pheromone_update', None) is not None:
            assert isinstance(optional_args['pheromone_update'], str) or \
                   isinstance(optional_args['pheromone_update'], dict), \
                'The pheromone_update parameter must be provided as a string or a dictionary as ' \
                'specified in the documentation.'
            if isinstance(optional_args['pheromone_update'], str):
                assert optional_args['pheromone_update'].lower() in PHEROMONE_UPDATE_STRATEGIES, \
                    'Pheromone update strategy %s not recognised. Available strategies: %r' % \
                    (optional_args['pheromone_update'], PHEROMONE_UPDATE_STRATEGIES)
            else:
                if optional_args['pheromone_update'].get('bag_size', None) is not None:
                    _intType('pheromone_update "bag_size"', optional_args['pheromone_update']['bag_size'],
                              min_val=1, max_val=args['n_ants'])
                assert 'strategy' in optional_args['pheromone_update'], \
                    'Hormone update strategy not defined using the "strategy" key. Available strategies: %r' % \
                    PHEROMONE_UPDATE_STRATEGIES
                assert optional_args['pheromone_update']['strategy'].lower() in PHEROMONE_UPDATE_STRATEGIES, \
                    'Pheromone update strategy %s not recognised. Available strategies: %r' % \
                    (optional_args['pheromone_update']['strategy'], PHEROMONE_UPDATE_STRATEGIES)
        if optional_args.get('scaleScores', None) is not None:
            _checkType('scaleScores', optional_args['scaleScores'], ScoreScaler)
        if optional_args.get('seed', None) is not None:
            _intType('seed', optional_args['seed'], min_val=0)

    @staticmethod
    def _selectPheromoneUpdate(kwargs: dict):
        """ Method that selects the pheromone update strategy provided by the user. """
        unrecognised_strategy = \
            'Pheromone update strategy not recognized, please check the parameter ' \
            'pheromone_update. Available strategies: %r' % PHEROMONE_UPDATE_STRATEGIES

        # Default pheromone update strategy
        if kwargs.get('pheromone_update', None) is None:
            update_strategy_ = updateAS('D')
        elif isinstance(kwargs['pheromone_update'], str):
            if kwargs['pheromone_update'].lower() == 'as':
                update_strategy_ = updateAS('D')
            elif kwargs['pheromone_update'].lower() == 'mmas':
                update_strategy_ = updateMMAS('D')
            elif kwargs['pheromone_update'].lower() == 'acs':
                update_strategy_ = updateACS('D')
            else:
                assert False, unrecognised_strategy
        elif isinstance(kwargs['pheromone_update'], dict):
            if 'strategy' not in kwargs['pheromone_update']:
                assert False, 'The key "strategy" in charge of specifying the update strategy is ' \
                              'not isPresent in the pheromone_update parameter.'
            if kwargs['pheromone_update']['strategy'].lower() == 'as':
                update_strategy_ = updateAS(kwargs.get('graph_type', 'D'),
                                            kwargs['pheromone_update'].get('elite', False))
            elif kwargs['pheromone_update']['strategy'].lower() == 'mmas':
                update_strategy_ = updateMMAS(kwargs.get('graph_type', 'D'),
                                              kwargs['pheromone_update'].get('elite', False))
            elif kwargs['pheromone_update']['strategy'].lower() == 'acs':
                update_strategy_ = updateACS(kwargs.get('graph_type', 'D'))
            else:
                assert False, unrecognised_strategy
        else:
            assert False, 'Unrecognised error in colony.ACO._selectPheromoneUpdate()'

        return update_strategy_

    @staticmethod
    def _selectPheromoneUpdateKW(kwargs: dict):
        """ Method that selects the optional arguments associated with the pheromone update and
        returns them as a dictionary. """
        if kwargs.get('pheromone_update', None) is None:
            return {'weight': 1.0}

        if isinstance(kwargs.get('pheromone_update', None), str):
            if kwargs['pheromone_update'].lower() == 'as':
                return {'weight': 1.0}
            elif kwargs['pheromone_update'].lower() == 'mmas':
                return {'limits': (0, 1), 'weight': 1.0}
            elif kwargs['pheromone_update'].lower() == 'acs':
                return {'decay': 0.1, 'weight': 1.0}

        elif isinstance(kwargs.get('pheromone_update', None), dict):
            if kwargs['pheromone_update']['strategy'].lower() == 'as':
                if kwargs['pheromone_update'].get('elite', None) is not None:
                    return {'elite': kwargs['pheromone_update']['elite'],
                            'weight': kwargs['pheromone_update'].get('weight', 1.0)}
                return {'weight': 1.0}
            elif kwargs['pheromone_update']['strategy'].lower() == 'mmas':
                if kwargs['pheromone_update'].get('elite', None) is not None:
                    return {'limits': tuple(kwargs['pheromone_update'].get('limits', (0, 1))),
                            'elite': kwargs['pheromone_update']['elite'],
                            'weight': kwargs['pheromone_update'].get('weight', 1.0)}
                return {'limits': tuple(kwargs['pheromone_update'].get('limits', (0, 1))),
                        'weight': kwargs['pheromone_update'].get('weight', 1.0)}
            elif kwargs['pheromone_update']['strategy'].lower() == 'acs':
                return {'decay': kwargs['pheromone_update'].get('decay', 0.1),
                        'weight': kwargs['pheromone_update'].get('weight', 1.0)}
        else:
            assert False, 'Unrecognised pheromone_update in antco.colony.ACO._selectPheromoneUpdateKW()'

    @property
    def seed(self) -> int:
        """ Returns the selected random seed. """
        return self._seed

    @property
    def graph(self) -> np.ndarray:
        """ Returns the graph defined by an adjacency matrix. """
        return self._graph

    @graph.setter
    def graph(self, input_val: tuple):
        """ Select a new graph matrix, to prevent accidental changes this must be received as a
        tuple where the first element will be the new graph matrix and the second a boolean true
        value. """
        assert input_val[1], 'Incorrect graph assignation protocol.'

        self._graph = input_val[0]

    @property
    def iterations(self) -> int:
        """ Returns the number of iterations to be performed. """
        return self._iterations

    @property
    def n_jobs(self) -> int:
        """ Returns the number of processors used for the execution of the algorithm. """
        return self._n_jobs

    @property
    def ants(self) -> list:
        """ Returns the list of ants used by the algorithm. """
        return self._ants

    @property
    def heuristic(self) -> np.ndarray:
        """ Returns the heuristic matrix. """
        return self._H

    @heuristic.setter
    def heuristic(self, input_val: tuple):
        """ Select a new heuristic matrix, to prevent accidental changes this must be received as a
        tuple where the first element will be the new heuristic matrix and the second a boolean
        true value. """
        assert input_val[1], 'Incorrect heuristic assignation protocol.'

        self._H = input_val[0]

    @property
    def precompute_heuristic(self):
        return self._precompute_heuristic

    @property
    def pheromones(self) -> np.ndarray:
        """ Returns the initialized pheromone matrix. """
        return self._P

    @pheromones.setter
    def pheromones(self, input_val: tuple):
        """ Select a new pheromone matrix, to prevent accidental changes this must be received as
        a tuple where the first element will be the new pheromone matrix and the second a boolean
        true value. """
        assert input_val[1], 'Incorrect pheromone assignation protocol.'

        self._P = input_val[0]

    @property
    def alpha(self) -> float:
        """ Returns the alpha parameter. """
        return self._alpha

    @property
    def beta(self) -> float:
        """ Returns the beta parameter. """
        return self._beta

    @property
    def pheromone_update_kw(self) -> dict:
        """ Returns the optional parameters used by the pheromone update strategy. """
        return self._pheromone_update_kw

    @property
    def updatePheromones(self) -> callable:
        """ Returns the pheromone update strategy. """
        return self._updatePheromones

    @property
    def pher_init_val(self) -> float:
        """ Returns the pheromone initial value. """
        return self._pher_init

    @property
    def objectiveFunction(self) -> ObjectiveFunction:
        """ Returns the objectiveFunction function to be maximized. """
        return self._objectiveFunction

    @property
    def tol(self) -> float:
        """ Returns the tolerance criteria for early stopping. """
        return self._tol

    @property
    def fixed_positions(self) -> bool:
        """ Returns a boolean value indicating whether to place the ants in fixed positions or
        not."""
        return self._fixed

    @property
    def Q(self) -> float or None:
        """ Returns the probability of selecting the next move deterministically by selecting the 
        move to the node that has the highest probability."""
        return self._Q

    @property
    def R(self) -> float or None:
        """ Returns the probability of selecting the next move randomly without taking into account 
        the computation of the pheromone matrix and heuristics """
        return self._R

    @property
    def scaleScores(self) -> ScoreScaler:
        """ Specifies whether to scale scores to the range 0-1 using the Min-Max algorithm. """
        return self._scaleScores
    
    @property
    def accessory_node(self) -> int or None:
        """ Returns the position of the accessory node (for more information see antco.preproc
        module). """
        return self._accessory_node
    
    @accessory_node.setter
    def accessory_node(self, node: int):
        """ Select the accessory node. """
        assert isinstance(node, int), 'Accessory node must be an integer.'
        assert node == self._graph.shape[0], 'Accessory node out of graph.'

        self._accessory_node = node


def getRandomWalk(initial_position: int, current_path: np.ndarray, adjacency_matrix: np.ndarray,
                  heuristic: np.ndarray, pheromone: np.ndarray, alpha: float, max_lim: int,
                  Q: float or None, R: float or None) -> np.ndarray:
    """
    Function that given an array indicating the nodes traversed (path traversed by an ant), a
    binary adjacency matrix indicating the structure of the graph to be traversed and the
    parameters that regulate the stochastic choices that the ants will make when choosing their
    movements (alpha and beta parameters that regulates the influence of the pheromone and
    heuristic values on the decisions taken by the ants) returns a binary adjacency matrix
    indicating the path traversed by the ant.

    Parameters
    ----------
    initial_position: int
        Integer indicating the initial position of the ant.

    current_path: np.ndarray (nodes), dtype=np.int8
        Array with nodes visited by the ant. The current_path argument must include the initial
        position of the ant.

    adjacency_matrix: np.ndarray (nodes, nodes), dtype=np.int8
        Binary adjacency matrix defining the structure of the graph to be traversed.

    heuristic: np.ndarray (nodes, nodes), dtype=np.float64
        Heuristic information matrix used by the stochastic ant policy to decide the ant's
        movements.

    pheromone: np.ndarray (nodes, nodes), dtype=np.float64
        Pheromone information matrix used by the stochastic ant policy to decide the ant's
        movements. The parameters of this matrix will be updated throughout the interactions of
        the algorithm.

    alpha: float
        Parameter that reference the influence of pheromones when the ant makes a decision on the
        path through the walk being constructed.

    max_lim: int
        Maximum path length.

    Q: float, default=None
        Parameter that determines the probability of selecting the next move deterministically 
        by selecting the move to the node that has the highest probability. By default this 
        parameter will not be considered.

    R: float, default=None
        Parameter that determines the probability of selecting the next move randomly without 
        taking into account the computation of the pheromone matrix and heuristics. By default 
        this parameter will not be considered.


    Returns
    -------
    :np.ndarray (nodes)
        Array with the nodes visited by the ant arranged in the order in which they have been
        visited.
    """
    movements = getValidPaths(initial_position, current_path, adjacency_matrix)
    n_partial_solutions = 1

    # Add partial solutions to the current path as long as possible.
    while len(movements) > 0 and n_partial_solutions < max_lim:
    
        if len(movements) == 1:
            mov = movements[0]
    
        elif Q is not None and random.random() < Q:    # Deterministic selection of the move
            probs = stochasticAS(
                initial_position, np.array(movements), heuristic, pheromone, alpha)        
            mov = movements[np.argmax(probs)]

        elif R is not None and random.random() < R:   # Random selection of the move
            mov = random.choice(movements)
        else:  # Stochastic selection of the next move            
            probs = stochasticAS(
                initial_position, np.array(movements), heuristic, pheromone, alpha)

            mov = movements[rouletteWheel(probs, random.random())]

        current_path = np.append(current_path, mov)
        movements = getValidPaths(mov, current_path, adjacency_matrix)
        initial_position = mov
        n_partial_solutions += 1

    return current_path


def step(ant: Ant, adjacency_matrix: np.ndarray, heuristic: np.ndarray, pheromone: np.ndarray,
         alpha: float, beta: float, seed: int, exp_heuristic: bool = True, Q: float = None,
         R: float = None) -> Ant:
    """
    Basic step function which ensures that an ant makes a path around the graph. The steps carried
    out by this function include:

        1. Initial positioning of the ant in the graph to be explored.
        2. Calling the getRandomWalk function so that the ant traverses the network.
        3. Processing of the path returned by the function.


    Parameters
    ----------
    ant: Ant
        Ant. Important note: for parallelization reasons this function will not modify the internal
        state of the received ant, it will returns a new ant.

    adjacency_matrix: np.ndarray (nodes, nodes), dtype=np.int8
        Binary adjacency matrix that defines the structure of the graph to be covered.

    heuristic: np.ndarray (nodes, nodes), dtype=np.float64
        Heuristic information matrix used by stochastic politics to select the nodes towards which
        the ant will move.

    pheromone: np.ndarray (nodes, nodes), dtype=np.float64
        Pheromone information matrix used by stochastic politics to select the edges towards which
        the ant will move. This matrix will be optimised over successive iterations of the algorithm.

    alpha: float
        Parameter that reference the influence of pheromones when the ant makes a decision on the
        path through the walk being constructed.

    beta: float
        Analogous to the alpha parameter, the beta parameter reference the importance given
        to the heuristic information.

    seed: int
        Random seed.

    exp_heuristic: bool, default=False
        Boolean value indicating whether to exponentiate the heuristic matrix to beta or not. By
        default it will not be exponentiated. It will be assumed that it has been previously
        exponentiated. Set this parameter to True for problems where the values of the heuristic
        matrix are changing throughout the interactions of the algorithm.

    Q: float, default=None
        Parameter that determines the probability of selecting the next move deterministically 
        by selecting the move to the node that has the highest probability. By default this 
        parameter will not be considered.

    R: float, default=None
        Parameter that determines the probability of selecting the next move randomly without 
        taking into account the computation of the pheromone matrix and heuristics. By default 
        this parameter will not be considered.

    Returns
    -------
    :Ant
        Returns the ant that has travelled along the network.
    """
    if exp_heuristic:
        heuristic = np.power(heuristic, beta)

    random.seed(seed)
    np.random.seed(seed)

    new_ant = Ant(l_min=ant.min_length, l_max=ant.max_length, graph_type=ant.representation,
                  check_params=False)

    new_ant.initAdjMatrix(n_nodes=adjacency_matrix.shape[0])

    if ant.initial_position is None:  # Random initial position
        initial_position = randomInit(adjacency_matrix)
    else:
        initial_position = ant.initial_position

    new_ant.setInitialPosition(initial_position)

    ant_path = getRandomWalk(
        new_ant.initial_position, new_ant.visited_nodes, adjacency_matrix, heuristic, pheromone,
        alpha, new_ant.max_length, Q, R)

    new_ant.visited_nodes = ant_path

    return new_ant


def generatePaths(ants: list, graph: np.ndarray, H: np.ndarray, P: np.ndarray, alpha: float,
                  beta: float, Q: float, R: float, n_jobs: int, seeds: np.ndarray,
                  exp_heuristic: bool = True) -> list:
    """
    Function that performs the exploration of the network according to the values of the heuristic
    and pheromone matrix.

    Parameters
    ----------
    ants: list
        List of ant instances.

    graph: np.ndarray (nodes, nodes), dtype=np.int8
        Graph to be explored.

    H: np.ndarray (nodes, nodes), dtype=np.float64
        Heuristic information.

    P: np.ndarray (nodes, nodes), dtype=np.float64
        Pheromone information.

    alpha: float
        The alpha parameter reference the influence of pheromones when the ant makes a decision
        on the path through the walk being constructed.

    beta: float
        Analogous to the alpha parameter, the beta parameter reference the importance given
        to the heuristic information received in H.

    Q: float, default=None
        Parameter that determines the probability of selecting the next move deterministically
        by selecting the move to the node that has the highest probability. By default this
        parameter will not be considered.

    R: float, default=None
        Parameter that determines the probability of selecting the next move randomly without
        taking into account the computation of the pheromone matrix and heuristics. By default
        this parameter will not be considered.

    n_jobs: int
        Number of processes to run in parallel.

    seeds: np.ndarray
        Random seeds.

    exp_heuristic: bool, default=True
        Parameter indicating whether to exponentiate the heuristic matrix to the beta value. By
        default it will not be assumed that the exponentiation has been precomputed.

    Returns
    -------
    :list
        List of ant instances that have traversed the graph to be explored.
    """
    if n_jobs == 1:
        ants = [
            step(ant, graph, H, P, alpha, beta, exp_heuristic=exp_heuristic, Q=Q, R=R, seed=seed)
            for ant, seed in zip(ants, seeds)]
    else:
        ants = joblib.Parallel(n_jobs=n_jobs, backend='loky')(
            joblib.delayed(step)(
                ant, graph, H, P, alpha, beta, exp_heuristic=exp_heuristic, Q=Q, R=R, seed=seed)
            for ant, seed in zip(ants, seeds))

    return ants


def generatePathsACS(ants: list, graph: np.ndarray, H: np.ndarray, P: np.ndarray, alpha: float,
                     beta: float, decay: float, pher_init: float, Q: float,
                     exp_heuristic: bool = True) -> list:
    """
    Function that performs the exploration of the graph using the Ant Colony System strategy
    proposed in

        Dorigo, M., & Gambardella, L. M. (1997). Ant colony system: a cooperative learning
        approach to the traveling salesman problem. IEEE Transactions on evolutionary computation,
        1(1), 53-66.

    Parameters
    ----------
    ants: list
        List of ant instances.

    graph: np.ndarray (nodes, nodes), dtype=np.int8
        Graph to be explored.

    H: np.ndarray (nodes, nodes), dtype=np.float64
        Heuristic information.

    P: np.ndarray (nodes, nodes), dtype=np.float64
        Pheromone information.

    alpha: float
        The alpha parameter reference the influence of pheromones when the ant makes a decision
        on the path through the walk being constructed.

    beta: float
        Analogous to the alpha parameter, the beta parameter reference the importance given
        to the heuristic information received in H.

    decay: float
        Decay to be applied during the local update of the pheromone matrix values after an ant
        has made the tour. This parameter is used in the local pheromone update given by equation

            P[i,j] = (1 - decay) * P[i,j] + decay * pher_init

    pher_init: float
        Parameter involved in the local update of the pheromone matrix values according to the
        equation

            P[i,j] = (1 - decay) * P[i,j] + decay * pher_init

    Q: float, default=None
        Parameter that determines the probability of selecting the next move deterministically
        by selecting the move to the node that has the highest probability. By default this
        parameter will not be considered.

    exp_heuristic: bool, default=True
        Parameter indicating whether to exponentiate the heuristic matrix to the beta value. By
        default it will not be assumed that the exponentiation has been precomputed.

    Returns
    -------
    :list
        List of ant instances that have traversed the graph to be explored.

    """
    if exp_heuristic:
        H_beta = np.power(H, beta)
    else:
        H_beta = H

    new_ants = []
    for ant in ants:
        new_ant = Ant(l_min=ant.min_length, l_max=ant.max_length, graph_type=ant.representation,
                      check_params=False)

        new_ant.initAdjMatrix(n_nodes=graph.shape[0])
        init_pos = randomInit(graph) if new_ant.initial_position is None else new_ant.initial_position
        new_ant.setInitialPosition(init_pos)

        # Generate random walk
        new_ant.visited_nodes = getRandomWalk(
            initial_position=new_ant.initial_position, current_path=new_ant.visited_nodes,
            adjacency_matrix=graph, heuristic=H_beta, pheromone=P, alpha=alpha,
            max_lim=new_ant.max_length, Q=Q, R=None)

        # Local pheromone update
        if new_ant.representation == 'u':
            updateUndLocalPher(ant=new_ant, P=P, decay=decay, init_val=pher_init)
        else:
            updateDirLocalPher(ant=new_ant, P=P, decay=decay, init_val=pher_init)

        new_ants.append(new_ant)

    return new_ants


def evaluateAnts(ants: list, objectiveFunction: ObjectiveFunction) -> np.ndarray:
    """
    Function that performs the evaluation of the paths traversed by the ants using the defined cost
    function.

    Parameters
    ----------
    ants: list
        List of ant instances.

    objectiveFunction: antco.optim.ObjectiveFunction
        Subclass of antco.optim.ObjectiveFunction defined by the user. This function will be
        maximized.

    Returns
    -------
    :np.ndarray (len(ants)), dtype=np.float64
        Scores associated with the ants.
    """
    ant_scores = np.array([
            objectiveFunction((ant, idx)) for idx, ant in enumerate(ants)], dtype=np.float64)[:, 0]

    return ant_scores




# Module that implements the different strategies for updating pheromone values.
#
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
from .ant import Ant
from .c_pheromone import (
    updateUndAS,
    updateDirAS,
    updateUndMMAS,
    updateDirMMAS,
    updateUndEliteAS,
    updateDirEliteAS,
    updateUndEliteMMAS,
    updateDirEliteMMAS,
    updateUndACS,
    updateDirACS)


def updateAS(topology: str, elite: bool = False):
    """
    Function that returns the Ant System pheromone value update strategy optimised on the basis of
    the graph topology received as an argument.

    Parameters
    ----------
    topology: str
        Graph topology: 'directed' or 'D' and 'undirected' or 'U'.

    elite: bool (default False)
        Indicates whether to use only the best ants for the pheromone update.

    Returns
    -------
    :function
        Algorithm optimised to work with the type of topology specified.
    """
    assert isinstance(topology, str), 'topology argument must be a string.'

    topology_ = topology.lower()

    if topology_ == 'd' or topology_ == 'directed':
        if elite:
            return updateDirEliteAS
        return updateDirAS
    elif topology_ == 'u' or topology_ == 'undirected':
        if elite:
            return updateUndEliteAS
        return updateUndAS
    else:
        assert False, 'Unrecognised topology parameter. Available options are: "directed" or "D" ' \
                      'and "undirected" or "U".'


def updateMMAS(topology: str, elite: bool = False):
    """
    Function that returns the MIN-MAX Ant System pheromone value update strategy optimised on the
    basis of the graph topology received as an argument.

    Parameters
    ----------
    topology: str
        Graph topology: 'directed' or 'D' and 'undirected' or 'U'.

    elite: bool (default False)
        Indicates whether to use only the best ants for the pheromone update.

    Returns
    -------
    :function
        Algorithm optimised to work with the type of topology specified.
    """
    assert isinstance(topology, str), 'topology argument must be a string.'

    topology_ = topology.lower()

    if topology_ == 'd' or topology_ == 'directed':
        if elite:
            return updateDirEliteMMAS
        return updateDirMMAS
    elif topology_ == 'u' or topology_ == 'undirected':
        if elite:
            return updateUndEliteMMAS
            
        return updateUndMMAS
    else:
        assert False, 'Unrecognised topology parameter. Available options are: "directed" or "D" ' \
                      'and "undirected" or "U".'


def updateACS(topology: str):
    """
    Function that returns the Ant Colony System pheromone update strategy optimised for the
    graph representation to be explored.

    Parameters
    ----------
    topology: str
        Graph topology: 'directed' or 'D' and 'undirected' or 'U'.

    Returns
    -------
    :function
        Algorithm optimised to work with the type of topology specified.
    """
    assert isinstance(topology, str), 'topology argument must be a string.'

    topology_ = topology.lower()

    if topology_ == 'd' or topology_ == 'directed':
        return updateDirACS
    elif topology_ == 'u' or topology_ == 'undirected':
        return updateUndACS
    else:
        assert False, 'Unrecognised topology parameter. Available options are: "directed" or "D" ' \
                      'and "undirected" or "U".'


def updateDirLocalPher(ant: Ant, P: np.ndarray, decay: float, init_val: float):
    """
    Local pheromone update from Ant Colony System (ACS) based on

        Dorigo, M., Birattari, M., & Stutzle, T. (2006). Ant colony optimization. IEEE
        computational intelligence magazine, 1(4), 28-39.

    The local pheromone update is performed by all the ants after each construction step. Each ant
    applies

        P[i,j] = (1 - decay) * P[i,j] + decay * P[i,j]

    only to the last edge traversed.

    Parameters
    ----------
    ant: antco.Ant
        Ant instance.

    P: np.ndarray (nodes, nodes), dtype=np.float64
        Pheromone matrix to be updated.

    decay: float
        Decay value according to the equation presented in the Ant Colony System for performing
        the local pheromone update.

    init_val: float
        Pheromone initial value.

    Notes
    -----
    Function optimised for directed graphs.
    """
    i, j = ant.visited_nodes[-2:]
    P[i, j] = (1 - decay) * P[i, j] + decay * init_val


def updateUndLocalPher(ant: Ant, P: np.ndarray, decay: float, init_val: float):
    """
    Local pheromone update from Ant Colony System (ACS) based on

        Dorigo, M., Birattari, M., & Stutzle, T. (2006). Ant colony optimization. IEEE
        computational intelligence magazine, 1(4), 28-39.

    The local pheromone update is performed by all the ants after each construction step. Each ant
    applies

        P[i,j] = (1 - decay) * P[i,j] + decay * P[i,j]

    only to the last edge traversed.

    Parameters
    ----------
    ant: antco.Ant
        Ant instance.

    P: np.ndarray (nodes, nodes), dtype=np.float64
        Pheromone matrix to be updated.

    decay: float
        Decay value according to the equation presented in the Ant Colony System for performing
        the local pheromone update.

    init_val: float
        Pheromone initial value.

    Notes
    -----
    Function optimised for undirected graphs.
    """
    i, j = ant.visited_nodes[-2:]
    P[i, j] = (1 - decay) * P[i, j] + decay * init_val
    P[j, i] = (1 - decay) * P[j, i] + decay * init_val



import numpy as np
import random
from antco import minMaxScaling, rouletteWheel


def test_min_max():
    """ antco.utils.minMaxScaling() testing unit """
    def round(value):
        return np.round(value, decimals=4)

    matrix = np.array([
        [4, 3, 4, 0],
        [1, 2, 4, 2],
        [1, 5, 4, 6],
        [1, 5, 4, 6]], dtype=float)

    graph = np.array([
        [0, 1, 0, 1],
        [0, 0, 0, 1],
        [1, 0, 0, 1],
        [0, 0, 0, 0]], dtype = np.int8)

    numbers_to_scale = np.array([3, 0, 2, 1, 6])

    scaled_matrix = (matrix - np.min(numbers_to_scale)) / (np.max(numbers_to_scale) - np.min(numbers_to_scale))
    scaled_matrix_test = minMaxScaling(matrix, graph)

    assert round(scaled_matrix[0, 1]) == round(scaled_matrix_test[0, 1]) and \
           round(scaled_matrix[0, 3]) == round(scaled_matrix_test[0, 3]) and \
           round(scaled_matrix[2, 0]) == round(scaled_matrix_test[2, 0]) and \
           round(scaled_matrix[2, 3]) == round(scaled_matrix_test[2, 3]), \
        'FAILED TEST: antco.utils.minMaxScaling()'

    print('SUCCESSFUL TEST: antco.utils.minMaxScaling()')


def test_roulette_wheel():
    """ antco.utils.rouletteWheel() testing unit """
    np.random.seed(1997)
    probs = np.array([0.0, 1.0, 0.0, 0.0, 0.0])

    index = rouletteWheel(probs, random.random())
    assert index == 1, 'FAILED TEST: antco.utils.rouletteWheel()'

    probs = np.zeros(shape=1000)
    probs[-1] = 1.0

    index = rouletteWheel(probs, random.random())

    assert index == 999, 'FAILED TEST: antco.utils.rouletteWheel()'

    print('SUCCESSFUL TEST: antco.utils.rouletteWheel()')


def test():
    test_min_max()
    test_roulette_wheel()




# Module containing the elements necessary to serialise objects.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import pickle
import os
import traceback
from datetime import datetime


def serialize(obj: object, path: str, extension: str = None):
    """
    Method to serialize any type of object received as an argument.

    Parameters
    ----------
    obj: object
        Object to be serialized.

    path: str
        Path where the object will be saved.

    extension: str (default None)
        Extension to be added.
    """
    abs_path = os.path.abspath(path)

    while os.path.exists(abs_path):  # If exists update the name
        path_to_file = '/'.join(abs_path.split('/')[:-1])
        file_name = abs_path.split('/')[-1].split('.')[0]
        abs_path = '%s/%s_(%s)' % \
                   (path_to_file, file_name, datetime.now().strftime("%m-%d-%Y-%H-%M-%S"))

        print('\nFile: %s already exists, name changed to %s' % (path, abs_path), end='')
        if extension is not None:
            print('.%s\n' % extension)
        else:
            print()

    # Add extension
    if extension is not None:
        if not abs_path.endswith('.%s' % extension):
            abs_path += '.%s' % extension

    with open(abs_path, 'ab') as out:
        pickle.dump(obj, out)

    print('%s object serialised correctly' % type(obj))


def load(path: str) -> object:
    """
    Method to deserialize the object stored in the file received as argument.

    Parameters
    ----------
    :param path: str
        Path where the object was saved.
    """
    abs_path = os.path.abspath(path)
    assert os.path.exists(abs_path), 'File %s not found.' % abs_path
    try:
        with open(path, 'rb') as input_file:
            obj = pickle.load(input_file)

        return obj
    except pickle.UnpicklingError as e:
        print(traceback.format_exc(e))
        raise
    except (AttributeError, EOFError, ImportError, IndexError) as e:
        print(traceback.format_exc(e))
        raise
    except Exception as e:
        print(traceback.format_exc(e))
        raise



import numpy as np
from antco import getBranchingFactor


def test_directed_branching_factor():
    """ antco.computeMetrics.getBranchingFactor() (DIRECTED)"""
    np.random.seed(1997)

    n_nodes = 6
    lambda_val = 0.5

    pheromones = np.random.uniform(size=(n_nodes, n_nodes))

    adj_matrix = np.array([
        [0, 0, 0, 1, 1, 1],
        [1, 0, 0, 1, 1, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 1],
        [0, 0, 0, 0, 0, 1],
        [0, 1, 0, 0, 0, 0]], dtype=np.int8)

    val = getBranchingFactor(pheromones, adj_matrix, lambda_val)
    assert val == 7, 'FAILED TEST: antco.computeMetrics.getBranchingFactor() (directed)'

    print('SUCCESSFUL TEST: antco.computeMetrics.getBranchingFactor() (directed)')



def test_undirected_branching_factor():
    """ antco.computeMetrics.getBranchingFactor() (UNDIRECTED)"""
    np.random.seed(1997)

    n_nodes = 6
    lambda_val = 0.5

    pheromones = np.random.uniform(size=(n_nodes, n_nodes))

    adj_matrix = np.array([
        [0, 1, 0, 1, 1, 1],
        [1, 0, 0, 1, 1, 0],
        [0, 0, 0, 0, 0, 0],
        [1, 1, 0, 0, 1, 1],
        [1, 1, 0, 1, 0, 1],
        [1, 0, 0, 1, 1, 0]], dtype=np.int8)

    for i in range(adj_matrix.shape[0]):
        for j in range(i + 1, adj_matrix.shape[1]):
            pheromones[i, j] = pheromones[j, i]

    val = getBranchingFactor(pheromones, adj_matrix, lambda_val)
    assert val == 10, 'FAILED TEST: antco.computeMetrics.getBranchingFactor() (undirected)'

    print('SUCCESSFUL TEST: antco.computeMetrics.getBranchingFactor() (undirected)')

def test():
    test_directed_branching_factor()
    test_undirected_branching_factor()



# Module containing the visualisation tools necessary to explore the convergence of the executed
# algorithms
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.ticker import MaxNLocator
from .report import Report


def convergence(report: Report, **kwargs):
    """
    Function that displays the convergence using a antco.report.Report object.

    Parameters
    ----------
    report: antco.report.Report
        antco.report.Report instance returned by the antco.run() function.

    **kwargs
        figsize: tuple, default=(8, 5)
            Tuple indicating the size of the figure.

        title: str, default='Convergence'
            Plot title.

        alpha_grid: float, default=0.7
            Transparency of the grid lines of the plot.

        alpha_graph: float, default=0.2
            Transparency of the lines of the plot.

        save_plot: str, default=None
            File in which to save the generated graph, if no value is provided the graph will not
            be saved.

    Returns
    -------
    :matplotlib.pyplot.Fig
        Figure with convergence graph.
    """

    def _draw(ax_, params_: dict, alpha_: float, color_: str, label_: str, linestyle_: str, linewidth_: int):
        x = np.arange(len(params_))
        y = [np.mean(vals) for vals in params_.values()]
        ax_.plot(x, y, color=color_, label=label_, alpha=alpha_, linestyle=linestyle_, linewidth=linewidth_)

        return ax_

    # Check that the parameters necessary to represent convergence can be obtained.
    try:
        report.get('mean_cost')
    except Exception:
        raise Exception(
            'The Report instance does not have the "mean_cost" value, make sure you have saved the "mean_cost" value '
            'throughout the interactions of the algorithm using the method report.save("mean_cost").')
    try:
        report.get('max_cost')
    except Exception:
        raise Exception(
            'The Report instance does not have the "max_cost" value, make sure you have saved the "max_cost" value '
            'throughout the interactions of the algorithm using the method report.save("max_cost").')

    parameters = {
        'mean_cost': {'color': '#85C1E9', 'label': 'Average cost', 'linestyle': 'solid', 'linewidth': 3},
        'max_cost': {'color': '#AF7AC5', 'label': 'Max cost', 'linestyle': 'dashed', 'linewidth': 2}}

    # Get optional arguments                  
    figsize = kwargs.get('figsize', (8, 5))
    title = kwargs.get('title', 'Convergence')
    alpha_graph = kwargs.get('alpha_graph', 0.7)
    alpha_grid = kwargs.get('alpha_grid', 0.2)
    save_plot = kwargs.get('save_plot', None)

    fig, ax = plt.subplots(figsize=figsize)

    for param, values in parameters.items():
        ax = _draw(ax, report.get(param), alpha_graph,
                   values['color'], values['label'], values['linestyle'], values['linewidth'])

    ax.set_xlabel('Iteration')
    ax.set_ylabel('Cost')
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.grid(alpha=alpha_grid)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.legend(loc='upper center', bbox_to_anchor=(0.5, 0.1),
              fancybox=True, shadow=True, ncol=len(parameters))
    ax.set_title(title)

    if save_plot is not None:
        plt.savefig(save_plot, dpi=150)

    return fig


def branchingFactor(report: Report, **kwargs):
    """
    Method representing the evolution of the lambda-branching factor.

    Parameters
    ----------
    report: antco.Report
        antco.report.Report instance returned by the antco.run() function.

    **kwargs
        figsize: tuple, default=(8, 5)
            Tuple indicating the size of the figure.

        title: str, default='Convergence'
            Plot title.

        alpha_grid: float, default=0.7
            Transparency of the grid lines of the plot.

        alpha_graph: float, default=0.2
            Transparency of the lines of the plot.

        color: str, default='viridis'
            Matplotlib colormap used to render the graph.

        show_legend: str, default=True
            Parameter indicating whether to display the legend or not.

        save_plot: str, default=None
            File in which to save the generated graph, if no value is provided the graph will not
            be saved.

    :return: matplotlib.pyplot.Fig
    """

    def _draw(ax_, params_: dict, alpha_: float, color_: str, label_: str):
        x = np.arange(len(params_))
        y = [np.mean(vals) for vals in params_.values()]
        ax_.plot(x, y, color=color_, label=label_, alpha=alpha_)

        return ax_

    assert len(report.values) > 0, 'The Report instance has not registered any value.'

    # Get optional arguments
    figsize = kwargs.get('figsize', (8, 5))
    title = kwargs.get('title', r'$\lambda$-branching factor')
    color = kwargs.get('color', 'viridis')
    alpha_graph = kwargs.get('alpha_graph', 0.7)
    alpha_grid = kwargs.get('alpha_grid', 0.2)
    show_legend = kwargs.get('show_legend', True)
    save_plot = kwargs.get('save_plot', None)

    fig, ax = plt.subplots(figsize=figsize)

    # Get unique keys
    unique_keys = [
        key for key in list(report.values[1].keys()) if 'lambda' in key]

    cmap = plt.cm.get_cmap(color, len(unique_keys))

    for idx, key in enumerate(unique_keys):
        ax = _draw(ax, report.get(key), alpha_graph, cmap(idx), key)

    ax.set_xlabel('Iteration')
    ax.set_ylabel('Number of nodes')
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.grid(alpha=alpha_grid)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    if show_legend:
        ax.legend()
    ax.set_title(title)

    if save_plot is not None:
        plt.savefig(save_plot, dpi=150)

    return fig


def pheromoneConvergence(report: Report, output: str = 'PheromoneConvergence.mp4',
                         color: str = 'Oranges'):
    """
    Method that generates a video showing how the values of the pheromone matrix evolve throughout
    the iterations of the algorithm.

    Parameters
    ----------
    report: antco.Report
        antco.report.Report instance returned by the antco.run() function.

    output: str
        Name of the output file under which the video will be saved.

    color: str
        Matplotlib colormap used to represent the pheromone values.
    """
    # Check that the parameters necessary to represent convergence can be obtained.
    try:
        report.get('pheromones')
    except Exception:
        raise Exception(
            'The Report instance does not have the "pheromones" value, make sure you have saved '
            'the "pheromones" value throughout the interactions of the algorithm using the method '
            'report.save("pheromones").')

    pheromone_evolution = report.get('pheromones')

    # Get max and min value
    min_val, max_val = None, None
    for values in pheromone_evolution.values():
        p_min = np.min(values)
        p_max = np.max(values)
        if min_val is None or p_min < min_val:
            min_val = p_min
        if max_val is None or p_max > max_val:
            max_val = p_max

    cmap = plt.cm.get_cmap(color)

    fig, ax = plt.subplots()

    ims = []
    for it, values in pheromone_evolution.items():
        im = ax.imshow(values, vmin=min_val, vmax=max_val, cmap=cmap, animated=True)
        if it == 1:  # Initial frame
            ax.imshow(values, vmin=min_val, vmax=max_val, cmap=cmap)
        ims.append([im])

    ani = animation.ArtistAnimation(fig, ims, interval=50, blit=True, repeat_delay=1000)

    # Save animation
    ani.save(output)

    plt.show()



# Module containing the interface to follow for the implementation of metaheuristics to run hybrid
# Ant Colony based algorithms.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import joblib
import multiprocessing as mp
import numpy as np
from antco.optim import ObjectiveFunction
from abc import ABCMeta, abstractmethod


class MetaHeuristic(object):
    """
    Class defining the interface to follow to implement a meta-heuristic that can be coupled to ant
    colony optimisation (hybrid approaches).

    Parameters
    ----------
    objective_function: antco.optim.ObjectiveFunction
        Objective function.

    add_to_old: bool, default=False
        Indicates whether to add the ants given by the hybrid to the ants received by the
        ant optimisation algorithm. By default only the ants returned by the hybrid
        strategy will be returned to the ACO algorithm for pheromone update.

    Abstract methods
    -------
    optimise(ants: list) -> list
         Method that receives a list of ants (whose attributes can be accessed through the methods
         defined for the antco.Ant class), applies the hybrid optimisation strategy
         implemented by the user and must return a tuple where the first element must correspond to
         a list of ants whose solutions must have been modified through the hybrid strategy, and
         the second element must correspond to the scores (in the same order) associated with the
         ants provided in the first parameter.
    """
    __metaclass__ = ABCMeta

    def __init__(self, objective_function: ObjectiveFunction, add_to_old: bool = False):
        assert isinstance(objective_function, ObjectiveFunction), \
            'objective_function must be a subclass of antco.optim.ObjectiveFunction'

        self.evaluate = objective_function
        self.__add_to_old = add_to_old

    def __repr__(self):
        return 'MetaHeuristic'

    def __str__(self):
        return self.__repr__()

    def __call__(self, ants: list) -> tuple:
        """
        Parameters
        ----------
        ants: list
           List of antco.Ant instances.

        Returns
        -------
        :tuple
            Returns a tuple where the first element corresponds to a list of ants and the second to
            the scores associated with each ant.
        """

        # Initial evaluation of the ants
        scores = self._evaluate(ants)

        # Apply hybrid strategy
        improved_ants, improved_scores = self.optimise(ants, scores)

        if self.evaluate.accessory_node:  # If an accessory node is being used, add to all ants
            init_pos = [ants[0].initial_position]
            for ant in improved_ants:
                ant.visited_nodes = init_pos + ant.visited_nodes.tolist()

        # Merge improved ants and old ants
        if self.__add_to_old:
            return ants + improved_ants, np.append(scores, improved_scores)

        return improved_ants, improved_scores

    def _evaluate(self, ants: list):
        """ Method to assess the quality of the received ant list using the defined objectiveFunction
        function."""
        assert len(ants) > 0, 'At least one ant must be passed to the hybrid'
        scores = np.array([
            self.evaluate((ant, idx)) for idx, ant in enumerate(ants)], dtype=np.float64)[:, 0]

        return scores

    @abstractmethod
    def optimise(self, ants: list, scores: list) -> tuple:
        """
        Method that using the paths travelled by the ants must return a list with an arbitrary
        number of ants (note that the number of ants returned does not have to coincide with the
        number of ants received).

        Parameters
        ----------
        ants: list
           List of antco.Ant instances.

        scores: list
            List of scores associated with each ant.

        Returns
        -------
        :tuple
            Returns a tuple where the first element corresponds to the list of improved ants and
            the second element to the scores given by the cost function. Note that if the evaluation
            procedure of the cost function defined in the hybrid strategy does not coincide with
            the cost function defined in the ACO algorithm, strange behaviours may occur.
        """
        raise NotImplementedError



# Module containing the necessary elements to monitor the evolution of the algorithms.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
from collections import defaultdict
from .c_metrics import getBranchingFactor
from .ant import Ant


class Report(object):
    """
    Object used to monitor the performance of a given algorithm.

    Methods
    -------
    computeMetrics(iteration: int, pheromones: np.ndarray, adj_matrix: np.ndarray, name: str)
    get(keyword: str): dict
    save(iteration: int, **kwargs)
    updateBest(iteration: int, score: float, ant: antco.Ant)


    Attributes
    ----------
    values: dict
    best_solution: dict

    Example
    -------
    >>> import antco
    >>>
    >>> report = antco.report.Report()

    To save parameters:

    >>> report.save(1, cost=1.0, performance=3.0)
    >>> report.save(2, cost=0.5, performance=4.0)
    >>> report.save(3, cost=0.1, performance=5.0)

    To retrieve the data:

    >>> report.get('cost')
    In[0]: {1: 1.0, 2: 0.5, 3: 0.1}

    >>> report.values
    In[1]: defaultdict(dict,
                      {1: {'cost': 1.0, 'performance': 3.0},
                       2: {'cost': 0.5, 'performance': 4.0},
                       3: {'cost': 0.1, 'performance': 5.0}})

    With these objects it is possible to monitor other aspects of the convergence of the algorithm,
    for example the lambda-branching factor

    >>> report = Report({
    >>>             'getBranchingFactor': [{'lambda_values': [0.05, 0.1]}]
    >>>                 })

    In the above example the lambda-branching factor will be monitored for lambda values 0.05 and
    0.1.

    To see which monitoring computeMetrics are implemented use:

    >>> help(antco.report.AVAILABLE_METRICS)

    """
    def __init__(self, metrics: dict = None):
        self._metrics = {}

        if metrics is not None:
            for metric, metric_args in metrics.items():
                assert metric.lower() in AVAILABLE_METRICS, 'Metric %s not available.' % metric
                self._metrics[METRICS[metric.lower()]] = metric_args

        self._values = defaultdict(dict)
        self._best_solution = {'solution': None, 'score': None, 'iteration': None, 'ant': None}

    def __repr__(self):
        return f'Report'

    def __str__(self):
        return self.__repr__()

    def __len__(self):
        return len(self._values)

    @property
    def values(self) -> dict:
        """
        Returns a dictionary with all the values stored during the execution of the algorithm.

        Returns
        -------
        :dict
            :key int
                Iteration number.
        """
        return self._values

    @property
    def best_solution(self) -> dict:
        """
        Return a dictionary with the best solution found by the algorithm.

        Returns
        -------
        :return dict
            :key "ant": Best Ant instance.
            :key "solution": Nodes visited by the ant.
            :key "score": Score associated to the ant.
            :key "iteration": Iteration in which the best solution was found.
        """
        return self._best_solution

    def computeMetrics(self, iteration: int, pheromones: np.ndarray, adj_matrix: np.ndarray,
                       name: str = None):
        """ 
        Function that computes the computeMetrics specified in the constructor.

        Parameters
        ----------
        iteration: int
            Current iteration.

        name: str
            Name preceding the names returned by the particular metric. For more information about
            the computeMetrics use help(antco.report.METRICS[METRIC_NAME]). To see the name of the computeMetrics
            that are available use help(antco.report.AVAILABLE_METRICS).

        pheromones: np.ndarray of shape (n_nodes, n_nodes), dtype=np.float64
            Pheromone matrix to be analysed.

        adj_matrix: np.ndarray of shape (n_nodes, n_nodes), dtype=np.int8
            Adjacency matrix defining the graph structure.
        """
        if len(self._metrics) > 0:
            for metric, metric_args in self._metrics.items():
                values = metric(pheromones, adj_matrix, **metric_args)

                self.save(iteration,  **{'%s_%s' % (name, key) if name is not None else key: value
                                         for key, value in values.items()})

    def get(self, keyword: str) -> dict:
        """
        Returns a dictionary where the key corresponds to the interaction number of the algorithm
        and the value to the value selected in set_cost for the argument received as parameter.

        Parameters
        ----------
        keyword: str
            Parameter from which to obtain the values saved during the execution of the algorithm.

        Returns
        -------
        :dict
            Dictionary where the key corresponds to the interaction number and the values to the
            data saved for that parameter.
         """
        if len(self._values.values()) == 0:
            assert False, 'The report object has no recorded value. For more information on how it ' \
                          'works use help(antco.report.Report).'

        assert keyword in list(self._values.values())[1], \
            'Parameter %s not found in Report. Available parameters: %r' % (keyword, list(self._values.values())[0].keys())

        return {iteration: values[keyword] for iteration, values in self._values.items()}

    def save(self, iteration: int, **kwargs):
        """
        Stores the values received as an argument associated with the interaction passed as an
        argument.

        Parameters
        ----------
        iteration: int
            Algorithm interaction number.

        kwargs:
            Parameters to be saved.
        """
        for key, value in kwargs.items():
            self._values[iteration][key] = value

    def updateBest(self, iteration: int, score: float, ant: Ant):
        """
        Method that saves the best solution found.

        Parameters
        ----------
        iteration: int
            Iteration.

        score: float
            Ant score.

        ant: Ant
            Ant instance.
        """
        self._best_solution['iteration'] = iteration
        self._best_solution['score'] = score
        self._best_solution['solution'] = ant.visited_nodes
        self._best_solution['ant'] = ant


def computeBranchingFactor(pheromones: np.ndarray, adj_matrix: np.ndarray, lambda_values: list = None,
                           **kw) -> dict:
    """ 
    Function that computes the lambda-branching factor for the specified lambda values. 

    Parameters
    ----------
    pheromones: np.ndarray of shape (n_nodes, n_nodes) and dtype np.float64
        Pheromone matrix to be analysed.

    adj_matrix: np.ndarray of shape (n_nodes, n_nodes) and dtype np.int8
        Adjacency matrix defining the graph structure.

    lambda_values: list
        List of lambda values to be computed.

    Returns
    -------
    :dict
        Dictionary where the keys will have the format 'lambda_VALUE_DE_LAMBDA' (i.e. lambda_0.05) and t
        he values will be the value obtained for the given lambda.

    """
    # Default lambda parameters
    lambda_values = lambda_values if lambda_values is not None else [0.05, 0.1, 0.2, 0.4]
    return {'lambda_%.4f' % val: getBranchingFactor(pheromones, adj_matrix, val) for val in lambda_values}


AVAILABLE_METRICS = ['branchingfactor']
METRICS = {
    AVAILABLE_METRICS[0]: computeBranchingFactor
}



# Module containing the main Ant-Colony Optimisation algorithms implemented.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
import sys
import random
import warnings
from copy import deepcopy
from .hybrid.base import MetaHeuristic
from .report import Report
from .ant import fixedPositions, deleteInitialPosition
from .base import DecaySchedule
from .aco import ACO, evaluateAnts, generatePaths

warnings.filterwarnings('error')


def updateReportWithBest(ants: list, scores: np.ndarray, best_score: float, report: Report,
                         iteration: int) -> float:
    """
    Function that updates the best solution of the report instance.

    Parameters
    ----------
    ants: list
        List of Ant instances.

    scores: np.ndarray, (len(ants)), dtype=np.float64
        Scores associated with the ants.

    best_score: float
        Best score seen until the current iteration.

    report: Report
        Report instance to be updated.

    iteration: int
        Current iteration.

    Returns
    -------
    :float
        Updated best score value.

    Notes
    -----
    This function will modify the internal state of the Report instance adding the best ant.
    """
    max_score = np.max(scores)

    # Update best score
    if max_score > best_score:
        best_score = max_score
        idx = np.argmax(scores)
        report.updateBest(iteration, best_score, deepcopy(ants[idx]))

    return best_score


def basic(aco_obj: ACO, metaheuristic: MetaHeuristic = None, apply_meta_each: int = 1,
          scores_decay: DecaySchedule = None, evaporation_decay: DecaySchedule = None,
          report: Report = None, save_pheromones: bool = True, verbose: bool = True) -> Report:
    """
    Function that executes a simple Ant Colony Optimization algorithm based on the parameters
    defined in the ACO object.

    Parameters
    ----------
    aco_obj: antco.ACO
        Initialised antco.ACO instance, for more information on this type of object use:
        help(antco.ACO)

    metaheuristic: antco.hybrid.base.MetaHeuristic subclass, default=None
        Instance following the interface defined in antco.extensions.MetaHeuristic. For more info
        use help(antco.extensions.MetaHeuristic).

    apply_meta_each: int, default=1
        Parameter indicating how many generations the hybrid will be applied for the
        refinement of the solutions given by the ants.

    scores_decay: antco.optim.DecaySchedule subclass, default=None
        Instance of a antco.optim.DecaySchedule representing a decay schedule that will applied to
        the score values with which the pheromone values are updated in each iteration.
        For more info use help(antco.optim.DecaySchedule).

    evaporation_decay: antco.optim.DecaySchedule subclass, default=None
        Same as scores_decay, in this case used to update the evaporation rate value.
        For more info use help(antco.optim.DecaySchedule).

    report: antco.report.Report, default=None
        antco.report.Report instance, by default a report will be initialised.

    save_pheromones: bool, default=True
        Value indicating whether to store in the report the values of the pheromone matrix at each
        iteration. If True these may be retrieved by:

            >>> report.get('pheromones')

    verbose: bool, default=True
        Indicates whether to show convergence information during the execution of the algorithm.

    Returns
    -------
    :antco.report.Report
        Returns a antco.report.Report instance with information about the convergence of the
        algorithm. The convergence parameters can be accessed via:

        >>> report = antco.algorithm.basic(...)
        >>> report.get('min_cost')
        >>> report.get('mean_cost')
        >>> report.get('max_cost')

        For more info about Report instances use help(antco.report.Report).
    """
    # Get parameters from aco_obj
    seed = aco_obj.seed
    graph = aco_obj.graph
    iterations = aco_obj.iterations
    n_jobs = aco_obj.n_jobs
    ants = aco_obj.ants
    H = aco_obj.heuristic
    P = aco_obj.pheromones
    alpha = aco_obj.alpha
    beta = aco_obj.beta
    if aco_obj.precompute_heuristic:
        H = np.power(H, beta)
    pheromone_update_kw = aco_obj.pheromone_update_kw
    rho = pheromone_update_kw['rho'];
    del pheromone_update_kw['rho']  # Get evaporation parameter
    updatePheromones = aco_obj.updatePheromones
    objectiveFunction = aco_obj.objectiveFunction
    tol = aco_obj.tol
    fixed = aco_obj.fixed_positions
    accessory_node = aco_obj.accessory_node
    Q = aco_obj.Q
    R = aco_obj.R
    scaleScores = aco_obj.scaleScores

    if report is None:  # Initialize a Report instance (if not provided)
        report = Report({'BranchingFactor': {'lambda_values': [0.5]}})

    if seed is not None:  # Set random state
        random.seed(seed)
        np.random.seed(seed)

    if fixed and accessory_node is None:  # Put ants in fixed positions (if specified)
        fixedPositions(ants, graph)

    # If an accessory node has been specified, place the ants on that node
    if accessory_node is not None:
        for ant in ants:
            ant.setInitialPosition(accessory_node)

    # Pre-compute the array to store the ant scores
    ant_scores, norm_scores = np.empty(shape=len(ants)), np.empty(shape=len(ants))

    current_iteration = 1
    it_without_improvements = 0
    best_score = -float('inf')
    while current_iteration <= iterations:
        seeds = np.random.randint(np.iinfo(np.int32).max, size=len(ants))
        ants = generatePaths(  # Graph exploration
            ants=ants, graph=graph, H=H, P=P, alpha=alpha, beta=beta, Q=Q, R=R, n_jobs=n_jobs,
            exp_heuristic=False, seeds=seeds)

        if metaheuristic is None or current_iteration % apply_meta_each != 0:
            # Evaluate ant paths using the objectiveFunction function (it will be maximized)
            ant_scores = evaluateAnts(ants=ants, objectiveFunction=objectiveFunction)
        else:
            ants, ant_scores = metaheuristic(ants)

        # Update best score and save best solution
        new_best = updateReportWithBest(
            ants=ants, scores=ant_scores, best_score=best_score, report=report,
            iteration=current_iteration)

        if new_best > best_score:
            best_score = new_best
            it_without_improvements = 0
        else:
            it_without_improvements += 1

        # Scale scores
        if scaleScores is not None:
            norm_scores = deepcopy(ant_scores)  # Copy scores
            norm_scores = scaleScores(norm_scores, best_score)
        else:
            norm_scores = ant_scores

        # Update pheromones according to the scores
        updatePheromones(
            paths=np.array([ant.adj_matrix for ant in ants], dtype=np.int8), P=P,
            ant_scores=norm_scores if scores_decay is None else norm_scores * scores_decay(
                current_iteration),
            rho=rho if evaporation_decay is None else rho * evaporation_decay(current_iteration),
            **pheromone_update_kw)

        # Compute convergence statistics
        mean_scores = np.mean(ant_scores)
        min_score = np.min(ant_scores)
        max_score = np.max(ant_scores)
        report.save(current_iteration, mean_cost=mean_scores, max_cost=max_score)

        if save_pheromones:  # Save pheromone values
            report.save(current_iteration, pheromones=deepcopy(P))

        # Compute monitoring computeMetrics
        report.computeMetrics(current_iteration, P, graph)

        if metaheuristic is None:
            if not fixed and accessory_node is None:  # Restart ants initial position
                deleteInitialPosition(ants)
        else:
            # Reposition the ants
            ants = aco_obj.ants
            if fixed and accessory_node is None:  # Put ants in fixed positions (if specified)
                fixedPositions(ants, graph)

            # If an accessory node has been specified, place the ants on that node
            if accessory_node is not None:
                for ant in ants:
                    ant.setInitialPosition(accessory_node)

        # After several generations without improvements, do an early stopping of the algorithm.
        if it_without_improvements > tol: break

        if verbose:
            sys.stdout.write('\rCost: Mean: %.4f (Min: %.4f Max: %.4f) (iteration %d)' %
                             (float(mean_scores), min_score, max_score, current_iteration))

        current_iteration += 1

    return report


def bagOfAnts(aco_obj: ACO, bag_size: int, out_of_bag_size: int = 0,
              metaheuristic: MetaHeuristic = None, apply_meta_each: int = 1,
              scores_decay: DecaySchedule = None, evaporation_decay: DecaySchedule = None,
              report: Report = None, save_pheromones: bool = True, verbose: bool = True) -> Report:
    """
    Function that executes an Ant Colony Optimization algorithm based on the parameters defined in
    the ACO object. This version of the algorithm maintains an bag_size aka a bag of ants (of the size
    specified in the execution parameters) that is preserved with the best ants. This bag will be
    the one used to update the values of the pheromone matrix (therefore, note that if an elitist
    strategy is used to update the values of the pheromone matrix, the size of the bag_size passed
    as an argument to the function and of the bag_size used to update the values of the pheromone
    matrix must be congruent).

    Parameters
    ----------
    aco_obj: antco.ACO
        Initialised antco.ACO instance, for more information on this type of object use:
        help(antco.ACO)

    bag_size: int
        Elite size maintained and used to update pheromone matrix values.

    out_of_bag_size: int, default=0
        Number of non-elite ants from the current iteration to include in the bag of ants.

    metaheuristic: antco.hybrid.base.MetaHeuristic subclass, default=None
        Instance following the interface defined in antco.extensions.MetaHeuristic. For more info
        use help(antco.extensions.MetaHeuristic).

    apply_meta_each: int, default=1
        Parameter indicating how many generations the hybrid will be applied for the
        refinement of the solutions given by the ants.

    scores_decay: antco.optim.DecaySchedule subclass, default=None
        Instance of a antco.optim.DecaySchedule representing a decay schedule that will applied to
        the score values with which the pher200omone values are updated in each iteration.
        For more info use help(antco.optim.DecaySchedule).

    evaporation_decay: antco.optim.DecaySchedule subclass, default=None
        Same as scores_decay, in this case used to update the evaporation rate value.
        For more info use help(antco.optim.DecaySchedule).

    report: antco.report.Report, default=None
        antco.report.Report instance, by default a report will be initialised.

    save_pheromones: bool, default=True
        Value indicating whether to store in the report the values of the pheromone matrix at each
        iteration. If True these may be retrieved by:

            >>> report.get('pheromones')

    verbose: bool, default=True
        Indicates whether to show convergence information during the execution of the algorithm.

    Returns
    -------
    :antco.report.Report
        Returns a antco.report.Report instance with information about the convergence of the
        algorithm. The convergence parameters can be accessed via:

        >>> report = bagOfAnts(...)
        >>> report.get('min_cost')
        >>> report.get('mean_cost')
        >>> report.get('max_cost')

        For more info about Report instances use help(antco.report.Report).
    """
    # Get parameters from aco_obj
    seed = aco_obj.seed
    graph = aco_obj.graph
    iterations = aco_obj.iterations
    n_jobs = aco_obj.n_jobs
    ants = aco_obj.ants
    H = aco_obj.heuristic
    P = aco_obj.pheromones
    alpha = aco_obj.alpha
    beta = aco_obj.beta
    if aco_obj.precompute_heuristic:
        H = np.power(H, beta)
    pheromone_update_kw = aco_obj.pheromone_update_kw
    rho = pheromone_update_kw['rho']; del pheromone_update_kw['rho']  # Get evaporation parameter
    updatePheromones = aco_obj.updatePheromones
    objectiveFunction = aco_obj.objectiveFunction
    tol = aco_obj.tol
    fixed = aco_obj.fixed_positions
    accessory_node = aco_obj.accessory_node
    Q = aco_obj.Q
    R = aco_obj.R
    scaleScores = aco_obj.scaleScores

    if report is None:  # Initialize a Report instance (if not provided)
        report = Report({'BranchingFactor': {'lambda_values': [0.5]}})

    if seed is not None:  # Set random state
        random.seed(seed)
        np.random.seed(seed)

    if fixed and accessory_node is None:  # Put ants in fixed positions (if specified)
        fixedPositions(ants, graph)

    # If an accessory node has been specified, place the ants on that node
    if accessory_node is not None:
        for ant in ants:
            ant.setInitialPosition(accessory_node)

    # Pre-compute the array to store the ant scores
    ant_scores, norm_scores = np.empty(shape=len(ants)), np.empty(shape=bag_size+out_of_bag_size)
    bag_of_ants = None
    boa_scores = np.full(shape=bag_size+out_of_bag_size, fill_value=-999_999.999)

    current_iteration = 1
    it_without_improvements = 0
    best_score = -float('inf')
    while current_iteration <= iterations:
        seeds = np.random.randint(np.iinfo(np.int32).max, size=len(ants))
        ants = generatePaths(   # Graph exploration
            ants=ants, graph=graph, H=H, P=P, alpha=alpha, beta=beta, Q=Q, R=R,
            n_jobs=n_jobs, exp_heuristic=False, seeds=seeds)

        if metaheuristic is None or current_iteration % apply_meta_each != 0:
            # Evaluate ant paths using the objectiveFunction function (it will be maximized)
            ant_scores = evaluateAnts(ants=ants, objectiveFunction=objectiveFunction)
        else:
            ants, ant_scores = metaheuristic(ants)
            assert len(ants) >= bag_size + out_of_bag_size, \
                'The number of ants returned by the metaheuristic strategy is lower than the ' \
                'number of ants that make up the bag of best ants. '

        # HACK: Elite upgrade (the current implementation is inefficient)
        if bag_of_ants is None:
            ordered_scores = np.argsort(ant_scores)[::-1]   # Ascending order
            bag_of_ants = [ants[ordered_scores[i]] for i in range(bag_size + out_of_bag_size)]
            boa_scores = np.array(ant_scores)[ordered_scores][:bag_size + out_of_bag_size]
            boa_scores = np.array(boa_scores)
        else:
            all_scores = np.append(boa_scores[:bag_size], ant_scores)
            all_ants = bag_of_ants[:bag_size] + ants
            ordered_scores = np.argsort(all_scores)[::-1]   # Ascending order
            bag_of_ants = [all_ants[ordered_scores[i]] for i in range(bag_size + out_of_bag_size)]
            boa_scores = np.array(all_scores)[ordered_scores][:bag_size + out_of_bag_size]

        # Update best score and save best solution
        new_best = updateReportWithBest(
            ants=bag_of_ants, scores=boa_scores, best_score=best_score, report=report,
            iteration=current_iteration)

        if new_best > best_score:
            best_score = new_best
            it_without_improvements = 0
        else:
            it_without_improvements += 1

        # Scale scores
        if scaleScores is not None:
            norm_scores[:] = boa_scores[:]  # Copy scores
            norm_scores = scaleScores(norm_scores, best_score)
        else:
            norm_scores = boa_scores

        # Update pheromones according to the scores
        updatePheromones(
            paths=np.array([ant.adj_matrix for ant in bag_of_ants], dtype=np.int8), P=P,
            ant_scores=norm_scores if scores_decay is None else norm_scores * scores_decay(current_iteration),
            rho=rho if evaporation_decay is None else rho * evaporation_decay(current_iteration),
            **pheromone_update_kw)

        # Compute convergence statistics
        mean_scores = np.mean(boa_scores)
        min_score = boa_scores[bag_size + out_of_bag_size - 1]
        max_score = boa_scores[0]
        report.save(current_iteration, mean_cost=mean_scores, max_cost=max_score)

        if save_pheromones:  # Save pheromone values
            report.save(current_iteration, pheromones=deepcopy(P))

        # Compute monitoring computeMetrics
        report.computeMetrics(current_iteration, P, graph)

        if metaheuristic is None:
            if not fixed and accessory_node is None:  # Restart ants initial position
                deleteInitialPosition(ants)
        else:
            # Reposition the ants
            ants = aco_obj.ants
            if fixed and accessory_node is None:  # Put ants in fixed positions (if specified)
                fixedPositions(ants, graph)

            # If an accessory node has been specified, place the ants on that node
            if accessory_node is not None:
                for ant in ants:
                    ant.setInitialPosition(accessory_node)

        # After several generations without improvements, do an early stopping of the algorithm.
        if it_without_improvements > tol: break

        if verbose:
            sys.stdout.write('\rCost: Mean: %.4f (Min: %.4f Max: %.4f) (iteration %d)' %
                             (float(mean_scores), min_score, max_score, current_iteration))

        current_iteration += 1

    return report




import numpy as np
from antco import (
    toDirAdjMatrix,
    toUndAdjMatrix,
    toDirAdjList,
    toUndAdjList,
    getValidPaths)


def test_to_directed_adjacency_matrix():
    """ antco.ntools.toDirAdjMatrix() testing unit """
    expected = [
        np.array([
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0]], dtype=np.int8),
        np.array([
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0]], dtype=np.int8)]

    arguments = [np.array([0, 1, 2, 4, 3]), np.array([0, 1, 2, 3])]

    for adj_matrix, adj_list in zip(expected, arguments):

        output = toDirAdjMatrix(adj_list, adj_matrix.shape[0])

        assert np.all(output == adj_matrix), 'FAILED TEST: antco.ntools.toDirAdjMatrix()'

    print('SUCCESSFUL TEST: antco.ntools.toDirAdjMatrix()')


def test_to_undirected_adjacency_matrix():
    """ antco.c_ntools.toUndAdjMatrix() testing unit """
    expected = [
        np.array([
            [0, 1, 0, 0, 0],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 0, 1],
            [0, 0, 0, 0, 1],
            [0, 0, 1, 1, 0]], dtype=np.int8),
        np.array([
            [0, 1, 0, 0, 0],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0]], dtype=np.int8)]

    arguments = [np.array([0, 1, 2, 4, 3]), np.array([0, 1, 2, 3])]

    for adj_matrix, adj_list in zip(expected, arguments):

        output = toUndAdjMatrix(adj_list, adj_matrix.shape[0])

        assert np.all(output == adj_matrix), 'FAILED TEST: antco.ntools.toUndAdjMatrix()'

    print('SUCCESSFUL TEST: antco.ntools.toUndAdjMatrix()')


def test_to_directed_adjacency_list():
    """ antco.ntools.toDirAdjList() testing unit """
    arguments = [
        (0, np.array([
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0]], dtype=np.int8)),
        (4, np.array([
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0],
            [1, 0, 0, 0, 0]], dtype=np.int8))]

    expected = [np.array([0, 1, 2, 4, 3]), np.array([4, 0, 1, 2, 3])]

    for args, adj_list in zip(arguments, expected):
        output = toDirAdjList(*args)

        assert np.all(output == adj_list), 'FAILED TEST: antco.ntools.toDirAdjList()'

    print('SUCCESSFUL TEST: antco.ntools.toDirAdjList()')


def test_to_undirected_adjacency_list():
    """ antco.ntools.toUndAdjList() testing unit """
    arguments = [
        (0, np.array([
            [0, 1, 0, 0, 0],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 0, 1],
            [0, 0, 0, 0, 1],
            [0, 0, 1, 1, 0]], dtype=np.int8)),
        (4, np.array([
            [0, 1, 0, 0, 1],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [1, 0, 0, 0, 0]], dtype=np.int8))]

    expected = [np.array([0, 1, 2, 4, 3]), np.array([4, 0, 1, 2])]

    for args, adj_list in zip(arguments, expected):
        output = toUndAdjList(*args)
        assert np.all(output == adj_list), 'FAILED TEST: antco.ntools.toUndAdjList()'

    print('SUCCESSFUL TEST: antco.ntools.toUndAdjList()')


def test_undirected_valid_paths():
    """ antco.ntools.getValidPaths() (undirected) testing unit """
    adj_matrix = np.array([
        [0, 1, 0, 1, 1, 0],
        [1, 0, 1, 1, 0, 0],
        [0, 1, 0, 0, 0, 1],
        [1, 1, 0, 0, 1, 1],
        [1, 0, 0, 1, 0, 0],
        [0, 0, 1, 1, 0, 0]], dtype=np.int8)

    # Initial position (0) -> Valid paths: (1, 3, 4)
    current_path = np.array([0], dtype=np.int64)
    choices = getValidPaths(0, current_path, adj_matrix)
    assert choices == [1, 3, 4], 'FAILED TEST: antco.ntools.test_valid_paths() (undirected)'

    # Initial position (2) -> Valid paths: (1, 5)
    current_path = np.array([0], dtype=np.int64)
    choices = getValidPaths(2, current_path, adj_matrix)
    assert choices == [1, 5], 'FAILED TEST: antco.ntools.test_valid_paths() (undirected)'

    # Initial position (2) -> Valid paths: (1) [5 visited]
    current_path = np.array([5], dtype=np.int64)
    choices = getValidPaths(2, current_path, adj_matrix)
    assert choices == [1], 'FAILED TEST: antco.ntools.test_valid_paths() (undirected)'

    print('SUCCESSFUL TEST: antco.ntools.test_valid_paths() (undirected)')


def test_directed_valid_paths():
    """ antco.ntools.getValidPaths() (directed) testing unit """
    # Initial position (0) -> Valid paths: (1)
    adj_matrix = np.array([
        [0, 1, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 1],
        [0, 0, 0, 0, 0, 0],
        [1, 0, 1, 0, 0, 0],
        [1, 0, 0, 1, 0, 0],
        [0, 0, 1, 0, 0, 0]], dtype=np.int8)

    current_path = np.array([0], dtype=np.int64)
    choices = getValidPaths(0, current_path, adj_matrix)
    assert choices == [1], 'FAILED TEST: antco.ntools.getValidPaths() (directed)'

    current_path = np.array([0], dtype=np.int64)
    choices = getValidPaths(1, current_path, adj_matrix)
    assert choices == [2, 5], 'FAILED TEST: antco.ntools.getValidPaths() (directed)'

    current_path = np.array([5], dtype=np.int64)
    choices = getValidPaths(1, current_path, adj_matrix)
    assert choices == [2], 'FAILED TEST: antco.ntools.getValidPaths() (directed)'

    current_path = np.array([5, 2], dtype=np.int64)
    choices = getValidPaths(1, current_path, adj_matrix)
    assert choices == [], 'FAILED TEST: antco.ntools.getValidPaths() (directed)'

    print('SUCCESSFUL TEST: antco.ntools.test_valid_paths() (directed)')


def test():
    test_to_directed_adjacency_matrix()
    test_to_undirected_adjacency_matrix()
    test_to_directed_adjacency_list()
    test_to_undirected_adjacency_list()
    test_undirected_valid_paths()
    test_directed_valid_paths()




import numpy as np
from antco import stochasticAS


def test_stochastic_policy():
    """ antco.policy.stochastic_policy() unit testing """
    np.random.seed(1997)

    # Test if function works
    movements = np.array([1, 2, 3, 4])
    H = np.random.uniform(size=(5, 5)).astype(np.float64)
    P = np.random.uniform(size=(5, 5)).astype(np.float64)
    alpha = 0.5
    _ = stochasticAS(0, movements, H, P, alpha)

    # Test function logic (movement should be from 0 to 2)
    H[:, :] = 0.0
    P[:, :] = 0.0
    H[0, 2] = 10.0
    P[0, 2] = 10.0

    probs = stochasticAS(0, movements, H, P, alpha)

    assert np.argmax(probs) == 1, 'FAILED TEST: antco.policy.stochastic_policy()'

    print('SUCCESSFUL TEST: antco.policy.stochastic_policy()')


def test():
    test_stochastic_policy()



# Module that defines the data structure representing the virtual ants of the algorithm and the
# functions applicable to this data structure.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
from .c_ntools import toDirAdjList, toUndAdjList, toDirAdjMatrix, toUndAdjMatrix


class Ant(object):
    """
    Object representing an ant.

    Parameters
    ----------
    l_min: int or float
        Minimum length of the path traversed by the ant for the solution found to be valid (i.e.
        is_valid to return True).

    l_max: int or float
        Maximum length of the path traversed by the ant for the solution found to be valid (i.e.
        is_valid to return True).

    graph_type: str, default='directed'
        Type of graph representation to be traversed by the ant. This parameter is necessary
        because depending on the type of the graph, different optimised algorithms will be used.
        Possible values:
            - 'd' or 'directed' for directed graphs.
            - 'u' or 'undirected' for undirected graphs.

    check_params: bool, default=True
        Parameter indicating whether to check the parameters received as arguments when creating
        an Ant instance. By default this value shall be set to True.

    Example
    --------
    >>> from antco import Ant, getValidPaths

    Create an ant that is going to operate on a directed network which, in order to be valid, must
    travel along paths containing from 1 to 10 edges.

    >>> ant_u = antco.Ant(1, 10, 'u')  # Ant for undirected graphs
    >>> ant = antco.Ant(1, 10, 'd')    # Ant for directed graphs
    >>> ant.is_valid
    In [0]: False
    >>> ant.max_length
    In [1]: 10
    >>> ant.min_length
    In [2]: 1
    >>> ant.representation
    In [3]: 'd'

    Initialise an empty binary adjacency matrix (for a graph of 10 nodes) to save the paths travelled.

    >>> ant.initAdjMatrix(n_nodes=10)

    Select the starting position of the ant.

    >>> ant.setInitialPosition(4)
    >>> ant.initial_position
    In [4]: 4

    Select initial position to a random (valid) position in a graph

    >>> adj_matrix = ...  # Binary adjacency matrix indicating the graph connections
    >>> ant.setInitialPosition(antco.randomInit(ant))
    >>> ant.initial_position
    In [5]: 2

    Distribute several ants along a network avoiding putting more than two ants in the same
    position as long as there are valid empty positions in the network.

    >>> ants = [antco.Ant(1, 10, 'u') for _ in range(10)]
    >>> # Initialise a densely connected undirected graph.
    >>> adj_matrix = np.ones(shape=(10, 10))
    >>> np.fill_diagonal(adj_matrix, 0)   # Fill diagonal with 0s
    >>> antco.fixedPositions(ants, adj_matrix)
    >>> # Print ant positions
    >>> ' - '.join(str(ant.initial_position) for ant in ants)
    In [6]: '8 - 2 - 3 - 0 - 5 - 7 - 1 - 6 - 4 - 9'

    Eliminate the initial positions of the ants.
    >>> antco.deleteInitialPosition(ants)
    >>> ' - '.join(str(ant.initial_position) for ant in ants)
    In [7]: 'None - None - None - None - None - None - None - None - None - None'

    Once the ant has traversed the graph (use help(antco.step) or see antco.step for more
    information) the list of nodes visited by the ant can be accessed using the attribute
    'visited_nodes':

    >>> ant.visited_nodes
    In [8]: array([3, 0, 0, 5, 6])
    >>> adj_matrix = ant.adj_matrix
    In [9]:
    array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 1, 0, 0, 0, 1, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
           [0, 0, 0, 0, 1, 1, 1, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]], dtype=int8)

    Creation of an ant from the values of the called ant.
    >>> new_ant = ant.new(initial_position=3)
    >>> new_ant.min_length
    In [10]: 1
    >>> new_ant.max_length
    In [11]: 10
    >>> new_ant.visited_nodes
    In [12]: array([])
    """
    def __init__(self, l_min: int or float, l_max: int or float, graph_type: str = 'directed',
                 check_params: bool = True):
        if check_params:
            assert (isinstance(l_min, int) or isinstance(l_min, float)) and \
                   (isinstance(l_max, int) or isinstance(l_max, float)), \
                'l_min and l_max must be integers or floats.'
            assert l_min < l_max, 'The lower limit cannot be equal to or less than the upper limit.'
            assert l_min >= 0, 'The lower limit cannot be less than 0.'
            assert isinstance(graph_type, str), 'graph_type must be a string.'

            graph_type_ = graph_type.lower()
            assert (graph_type_ == 'd') or (graph_type_ == 'u') or (graph_type_ == 'directed') or \
                   (graph_type_ == 'undirected'), \
                'The accepted values for the graph_type parameter are: "directed" or "d" for directed ' \
                'graphs and "undirected" or "u" for undirected graphs.'

        graph_type_ = graph_type.lower()

        # Select the mapping functions between adjacency lists and adjacency matrices
        if graph_type_ == 'd' or graph_type_ == 'directed':
            self._toAdjList = toDirAdjList
            self._toAdjMatrix = toDirAdjMatrix
            self._graph_type = 'd'
        else:
            self._toAdjList = toUndAdjList
            self._graph_type = 'u'
            self._toAdjMatrix = toUndAdjMatrix

        self._path_limits = (l_min, l_max)
        self._adjacency_list = None
        self._adjacency_matrix = None
        self._initial_position = None
        self._mapped = False
        self._positioned = False

    def __repr__(self):
        return f'Ant(path_limits={self._path_limits}, graph_type={self._graph_type})'

    def __str__(self):
        return self.__repr__()

    def initAdjMatrix(self, n_nodes: int):
        """
        Method that creates an empty binary adjacency matrix of the specified number of nodes.

        Parameters
        ----------
        n_nodes: int
            Number of nodes of the adjacency matrix.
        """
        self._adjacency_matrix = np.zeros(shape=(n_nodes, n_nodes), dtype=np.int8)
        np.fill_diagonal(self._adjacency_matrix, 1)
        self._mapped = False

    def setInitialPosition(self, position: int or None):
        """
        Method for selecting the initial position of the ant.

        Parameters
        ----------
        position: int
            Initial position.

        Notes
        -----
        For efficiency reasons the verification of the received values is disabled.
        """
        self._initial_position = position
        self._mapped = False
        self._positioned = True

    @property
    def initial_position(self) -> int:
        """ Method that returns the initial position of the ant. """
        return self._initial_position

    @property
    def visited_nodes(self) -> np.ndarray:
        """
        Returns an array with the visited nodes (the indices of the visited nodes, i.e. a list with
        the partial solutions).

        Returns
        -------
        :np.ndarray
            Array with the nodes visited by the ant.

        Notes
        -----
        Note the partial solutions found by an ant are internally represented as adjacent matrices.
        Calling this function transforms the adjacency matrix into an array. As long as the
        adjacency matrix is not modified the following calls to the function will use the previously
        computed array without the need to transform the adjacency matrix back into an array.
        """
        if self._adjacency_matrix is None or not self._positioned:
            return np.array([], dtype=np.int64)

        if self._mapped:
            return self._adjacency_list
        self._adjacency_list = self._toAdjList(self._initial_position, self._adjacency_matrix)
        self._mapped = True

        return self._adjacency_list

    @visited_nodes.setter
    def visited_nodes(self, adj_list):
        """ Method for assigning an adjacency list to an ant.  """
        assert len(adj_list) > 1, 'The adjacency list must have at least two nodes.'
        self.setInitialPosition(adj_list[0])
        self.adj_matrix = self._toAdjMatrix(np.array(adj_list, np.int64), self._adjacency_matrix.shape[0])

    @property
    def adj_matrix(self) -> np.ndarray or None:
        """
        Method that returns the binary adjacency matrix representing the path chosen by the ant. If
        the ant has no path the method will return None.

        Returns
        -------
        :np.ndarray (nodes, nodes) or None
             Binary adjacency matrix or None.
        """
        return self._adjacency_matrix

    @adj_matrix.setter
    def adj_matrix(self, new_adj_matrix: np.ndarray):
        """
        Setter that allows the selection of a new binary adjacency matrix for a given ant.

        Parameters
        ----------
        new_adj_matrix: np.ndarray (nodes, nodes), dtype=np.int8
            New integer square adjacency matrix where the positions with value 1 indicate the
            connection between the nodes indicated by the row and column indices.

        Notes
        -----
        For efficiency reasons this method will only check that the matrix received is square. IT
        WILL NOT CHECK THE TYPE OF DATA IN THE RECEIVED MATRIX.
        """
        assert len(new_adj_matrix.shape) == 2 and new_adj_matrix.shape[0] == new_adj_matrix.shape[1], \
            'The dimensions of the adjacency matrix received are not consistent, it must be a square ' \
            'matrix of integer type.'

        self._mapped = False
        self._adjacency_matrix = new_adj_matrix

    @property
    def max_length(self) -> int:
        """
        Method that informs about the maximum size allowed for the path taken by the ant.

        Returns
        -------
        :int
            Maximum path length.
        """
        return self._path_limits[1]

    @property
    def min_length(self) -> int:
        """
        Method that informs about the minimum size allowed for the path taken by the ant.

        Returns
        -------
        :int
            Minimum path length.
        """
        return self._path_limits[0]

    @property
    def is_valid(self) -> bool:
        """
        Method that informs about the minimum size restrictions allowed for a partial solution.

        Returns
        -------
        :bool
            True if the number of elements in the partial solution is within the permitted limits,
            otherwise False.
        """
        length = len(np.where(self._adjacency_matrix == 1)[0]) if self._graph_type == 'd' else \
                 len(np.where(self._adjacency_matrix == 1)[0]) // 2

        if self._path_limits[0] <= length <= self._path_limits[1]:
            return True

        return False

    @property
    def representation(self) -> str:
        """
        Returns the graph representation type that is to be walked through.

        Returns
        -------
        :str
            Representation: 'd' for directed graphs and 'u' for undirected.
        """
        return self._graph_type

    def new(self):
        """
        Method to construct an ant while preserving the l_min, l_max and graph_type values of the
        calling instance. This method will automatically handle the call to initAdjMatrix().

        Returns
        -------
        :Ant
            Instance of the Ant class with parameters l_min, l_max and graph_type equal to those of
            the calling instance.
        """
        new_ant = Ant(l_min=self.min_length, l_max=self.max_length, graph_type=self.representation)
        new_ant.initAdjMatrix(n_nodes=self.adj_matrix.shape[0])

        return new_ant


def randomInit(adjacency_matrix: np.ndarray) -> int:
    """
    Method that randomly selects a position within the binary adjacency matrix.

    Parameters
    ----------
    adjacency_matrix: np.ndarray of shape (n_nodes, n_nodes), dtype=np.int8
        Binary adjacency matrix that defines the structure of the network and whose diagonal
        should be 0.

    Returns
    -------
    :int
        Integer indicating the initial position of the ant on the graph selected at random.
    """
    available_indices = np.unique(np.where(adjacency_matrix == 1))

    return np.random.choice(available_indices)


def fixedPositions(ants: list, adjacency_matrix: np.ndarray) -> None:
    """
    Function that places ants in fixed positions at different positions in the adjacency matrix by
    distributing them uniformly across the graph.

    Parameters
    ----------
    ants: list
        List with ants that will be placed in fixed positions, their internal state will be modified.

    adjacency_matrix: np.ndarray of shape (n_nodes, n_nodes), dtype=np.int8
        Binary adjacency matrix that defines the structure of the network and whose diagonal should
        be 0.
    """
    available_positions = np.unique(np.array(np.where(adjacency_matrix == 1)))
    for ant in ants:
        i = np.random.randint(len(available_positions))
        ant.setInitialPosition(available_positions[i])
        # Eliminate indices where ants are positioned
        available_positions = np.delete(available_positions, i)
        # If all the indices are already filled with ants, start placing more ants on the visited indices
        if len(available_positions) == 0:
            available_positions = np.unique(np.array(np.where(adjacency_matrix == 1)))


def deleteInitialPosition(ants: list) -> None:
    """
    Method that selects the initial position of the list of ants to None.

    Parameters
    ----------
    ants: list
        List with ants that will be placed in fixed positions, their internal state will be modified.
    """
    for ant in ants:
        ant.setInitialPosition(None)



from .pheromone import updateAS, updateMMAS, updateDirLocalPher, updateUndLocalPher
from .ant import Ant, randomInit, fixedPositions, deleteInitialPosition
from .aco import *
from . import optim
from . import report
from . import graphics
from . import preproc
from . import serialize
from . import algorithm
from . import tools

# Cython modules
from .c_utils import minMaxScaling, rouletteWheel
from .c_policy import stochasticAS
from .c_metrics import getBranchingFactor
from .c_ntools import (
    toDirAdjMatrix,
    toUndAdjMatrix,
    toDirAdjList,
    toUndAdjList,
    getValidPaths)

from .c_pheromone import (
    updateUndAS,
    updateDirAS,
    updateUndMMAS,
    updateDirMMAS,
    updateDirEliteMMAS,
    updateUndEliteMMAS,
    updateDirEliteAS,
    updateUndEliteAS,
    updateUndACS,
    updateDirACS
)

# Experimental algorithms (development)
from .experimental import experimental



import numpy as np
import joblib
from antco import getRandomWalk, Ant, step


def test_undirected_random_walk():
    """ antco.colony.undirected_random_walk() unit testing"""
    np.random.seed(1997)

    heuristic = np.random.uniform(size=(6, 6)).astype(np.float64)
    pheromone = np.random.uniform(size=heuristic.shape).astype(np.float64)
    alpha, beta = 1.0, 1.0
    max_lim = 10
    Q = 0.0001
    R = 0.0001

    adj_matrix = np.array([
        [0, 1, 0, 1, 1, 0],
        [1, 0, 1, 0, 0, 1],
        [0, 1, 0, 1, 0, 1],
        [1, 0, 1, 0, 1, 0],
        [1, 0, 0, 1, 0, 0],
        [0, 1, 1, 0, 0, 0]], dtype=np.int8)

    possible_paths = [
        [0, 3, 2, 1, 5],
        [0, 3, 2, 5, 1],
        [0, 4, 3, 2, 5, 1],
        [0, 3, 4],
        [0, 4, 3, 2, 1, 5],
        [0, 1, 2, 5],
        [0, 1, 2, 3, 4],
        [0, 1, 5, 2, 3, 4]]

    for _ in range(1_000):
        current_path = np.array([0])
        path = getRandomWalk(initial_position=4,
                             current_path=current_path,
                             adjacency_matrix=adj_matrix,
                             heuristic=heuristic,
                             pheromone=pheromone,
                             alpha=alpha,
                             max_lim=max_lim,
                             Q=Q, R=R).tolist()

        assert path in possible_paths, 'ERROR: antco.colony.undirected_random_walk()'

    print('SUCCESSFUL TEST: antco.colony.undirected_random_walk()')


def test_directed_random_walk():
    """ antco.colony.directed_random_walk() unit testing """
    np.random.seed(1997)

    heuristic = np.random.uniform(size=(6, 6)).astype(np.float64)
    pheromone = np.random.uniform(size=heuristic.shape).astype(np.float64)
    alpha, beta = 1.0, 1.0
    max_lim = 10
    Q = 0.0001
    R = 0.0001

    # Initial position (4) -> Valid paths: [0, 3]
    adj_matrix = np.array([
        [0, 1, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 1],
        [0, 0, 0, 0, 0, 0],
        [1, 0, 1, 0, 0, 0],
        [1, 0, 0, 1, 0, 0],
        [0, 0, 1, 0, 0, 0]], dtype=np.int8)

    possible_paths = [
        [4, 3, 0, 1, 2],
        [4, 0, 1, 2],
        [4, 0, 1, 5, 2],
        [4, 3, 2],
        [4, 3, 0, 1, 5, 2]
    ]

    for _ in range(1_000):
        current_path = np.array([4], dtype=np.int64)
        path = getRandomWalk(initial_position=4,
                             current_path=current_path,
                             adjacency_matrix=adj_matrix,
                             heuristic=heuristic,
                             pheromone=pheromone,
                             alpha=alpha,
                             max_lim=max_lim,
                             Q=Q, R=R).tolist()
        assert path in possible_paths, 'ERROR: antco.colony.directed_random_walk()'

    print('SUCCESSFUL TEST: antco.colony.directed_random_walk()')


def undirected_step():
    """ UNDIRECTED:
            antco.colony.step()
            antco.colony.update_ants()
        unit testing.
    """
    np.random.seed(1997)

    graph_type = 'u'
    adjacency_matrix = np.ones(shape=(6, 6), dtype=np.int8)
    heuristic = np.random.uniform(size=adjacency_matrix.shape).astype(np.float64)
    pheromone = np.random.uniform(size=adjacency_matrix.shape).astype(np.float64)
    alpha, beta = 1.0, 1.0
    max_lim = 40
    n_ants = 3

    for n_jobs in range(1, 3):
        ants = [Ant(l_min=0, l_max=max_lim, graph_type=graph_type) for _ in range(n_ants)]
        seeds = np.random.randint(np.iinfo(np.int32).max, size=len(ants))

        if n_jobs == 1:
            paths = []
            for ant, seed in zip(ants, seeds):
                paths.append(step(
                    ant=ant, adjacency_matrix=adjacency_matrix, heuristic=heuristic,
                    pheromone=pheromone, alpha=alpha, beta=beta, exp_heuristic=True,
                    Q=0.0001, R=0.0001, seed=seed))
        else:
            paths = joblib.Parallel(n_jobs=n_jobs)(
                joblib.delayed(step)(
                    ant, adjacency_matrix, heuristic, pheromone, alpha, beta, True, 0.0001, 0.0001)
                for ant, seed in zip(ants, seeds))

    print('SUCCESSFUL TEST: antco.colony.step() [UNDIRECTED]')


def directed_step():
    """ DIRECTED:
            antco.colony.step()
            antco.colony.update_ants()
        unit testing.
    """
    np.random.seed(1997)

    graph_type = 'd'
    adjacency_matrix = np.ones(shape=(6, 6), dtype=np.int8)
    for _ in range(10):
        i, j = np.random.randint(6, size=2)
        adjacency_matrix[i, j] = 0

    heuristic = np.random.uniform(size=adjacency_matrix.shape).astype(np.float64)
    pheromone = np.random.uniform(size=adjacency_matrix.shape).astype(np.float64)
    alpha, beta = 1.0, 1.0
    max_lim = 40
    n_ants = 3

    for n_jobs in range(1, 3):
        ants = [Ant(l_min=0, l_max=max_lim, graph_type=graph_type) for _ in range(n_ants)]
        seeds = np.random.randint(np.iinfo(np.int32).max, size=len(ants))

        if n_jobs == 1:
            paths = []
            for ant, seed in zip(ants, seeds):
                paths.append(step(
                    ant=ant, adjacency_matrix=adjacency_matrix, heuristic=heuristic,
                    pheromone=pheromone, alpha=alpha, beta=beta, exp_heuristic=True,
                    Q=0.0001, R=0.0001, seed=seed))
        else:
            paths = joblib.Parallel(n_jobs=n_jobs)(
                joblib.delayed(step)(
                    ant=ant, adjacency_matrix=adjacency_matrix, heuristic=heuristic,
                    pheromone=pheromone, alpha=alpha, beta=beta, exp_heuristic=True,
                    Q=0.0001, R=0.0001, seed=seed)
                for ant, seed in zip(ants, seeds))

    print('SUCCESSFUL TEST: antco.colony.step() [DIRECTED]')


def test():
    test_directed_random_walk()
    test_undirected_random_walk()
    undirected_step()
    directed_step()



# Module that collects the interfaces that must follow certain components used in the framework.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator
from abc import ABCMeta, abstractmethod


class DecaySchedule(object):
    """
    Class representing a decay schedule that is applied to the score values or evaporation
    parameter which are used for update the pheromone values in each iteration.

    Methods
    -------
    show()
        Method that allows to represent how the weighting evolves as the iterations progress
        according to the definition given by the user of the abstract decay method.

    Abstract methods
    ----------------
    decay(iteration: int) -> float

    """
    __metaclass__ = ABCMeta

    def __repr__(self):
        return f'DecaySchedule'

    def __str__(self):
        return self.__repr__()

    def __call__(self, iteration: int) -> float:
        return self.decay(iteration)

    @abstractmethod
    def decay(self, iteration: int) -> float:
        """
        Method that must return a numeric value that will act as a decay on the scores or
        evaporation parameter. In this way the scores or evaporation will be multiplied by the
        value received. This allow scores or evaporation to be smoothed as a function of the
        current iteration. Depending on which iteration the algorithm is in the pheromones
        will be updated in a more abrupt or more smooth way depending on the value returned.

        Parameters
        ----------
        iteration: int
            Algorithm iteration.

        Returns
        -------
        :float
            Decay value
        """
        raise NotImplementedError

    def show(self, from_it: int = 1, to_it: int = 100, title: str = 'Decay schedule'):
        """
        Method that shows a plot of how the weighting applied by the decay evolves as a function
        of the number of interactions.

        Parameters
        ----------
        from_it: int, default=1
            Initial iteration.

        to_it: int, default=100
            Last iteration.

        title: str, default='Decay schedule'
            Plot title.
        """
        assert from_it > 0, 'Initial iteration (from_it parameter) must be > 1.'
        assert from_it < to_it, 'Initial iteration (from_it parameter) must be > final iteration (to_it parameter).'

        fig, ax = plt.subplots(figsize=(8, 5))

        x = np.arange(from_it, to_it)
        y = [self.decay(it) for it in x]

        ax.scatter(x, y, c=plt.cm.Oranges(y))
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.grid(alpha=0.2)
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.set_title(title)

        plt.show()


class ScoreScaler(object):
    """
    Class defining the interface followed by those classes used to scale the values of the scores
    given by the cost function to a given range.

    Abstract methods
    ----------------
    scale(ant_scores: np.ndarray, best_historic: float)
    """
    __metaclass__ = ABCMeta

    def __repr__(self):
        return f'ScoreScaler'

    def __str__(self):
        return self.__repr__()

    def __call__(self, ant_scores: np.ndarray, best_historic: float):
        return self.scale(ant_scores, best_historic)

    def scale(self, ant_scores: np.ndarray, best_historic: float):
        """
        Method that receives as arguments an array with the scores given by the cost function and
        the best score seen by the algorithm and must return an array of the same size as the one
        received with the scores scaled as decided by the user.

        Parameters
        ----------
        ant_scores: np.ndarray (n_ants), dtype=np.float64
            Scores given by the objective function.

        best_historic: float
            Best score found until the current iteration.

        Returns
        -------
        :np.ndarray (n_ants), dtype=np.float64
            Scaled scores.
        """
        raise NotImplementedError



import numpy as np
from copy import deepcopy
from antco import (
    updateUndAS,
    updateDirAS,
    updateUndMMAS,
    updateDirMMAS,
    updateUndEliteMMAS,
    updateDirEliteMMAS,
    updateDirEliteAS,
    updateUndEliteAS,
    updateDirLocalPher,
    updateUndLocalPher,
    updateUndACS,
    updateDirACS)

from antco import Ant


def test_directed_AS_update():
    """ antco.pheromone.updateDirAS() unit testing """
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T) / 2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 2
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 1],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)
    expected = np.array([
        [0.0, 0.9267931249792329, 0.4776117072586296, 1.6791352931971335],
        [0.9267931249792329, 0.0, 0.5591658434565883, 0.7150135839042728],
        [0.4776117072586296, 0.5591658434565883, 0.0, 1.0865920636193305],
        [1.6791352931971335, 0.7150135839042728, 1.0865920636193305, 0.0]], dtype=np.float64)

    ant_scores = np.array([0.2, 0.3, 0.4], dtype=np.float64)

    updateDirAS(paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, weight=1.0)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateDirAS()'

    print('SUCCESSFUL TEST: antco.pheromone.updateDirAS()')


def test_undirected_AS_update():
    """ antco.pheromone.updateUndAS() unit testing """
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T)/2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 2
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 1],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)
    expected = np.array([
        [0.0, 0.9267931249792329, 0.4776117072586296, 1.6791352931971335],
        [0.9267931249792329, 0.0, 0.5591658434565883, 0.7150135839042728],
        [0.4776117072586296, 0.5591658434565883, 0.0, 1.0865920636193305],
        [1.6791352931971335, 0.7150135839042728, 1.0865920636193305, 0.0]], dtype=np.float64)

    ant_scores = np.array([0.2, 0.3, 0.4]).astype(np.float64)

    updateUndAS(paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, weight=1.0)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateUndAS()'

    print('SUCCESSFUL TEST: antco.pheromone.updateUndAS()')


def test_directed_AS_elite_update():
    """ antco.pheromone.updateDirEliteAS() unit testing """
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T) / 2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [1, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 2
        [[0, 1, 1, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 1],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)

    expected = np.array([
        [0.0, 0.6414344987114436, 0.6820893643835099, 1.2433082310436099],
        [0.6414344987114436, 0.0, 0.4473326730988265, 0.5720108649925117],
        [0.3820893643835099, 0.4473326730988265, 0.0, 0.7692736491472838],
        [1.2433082310436099, 0.5720108649925117, 0.7692736491472838, 0.0]],
        dtype=np.float64)

    ant_scores = np.array([0.2, 0.3, 0.4], dtype=np.float64)

    updateDirEliteAS(
        paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, elite=2, weight=1.0)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.directed_AS_elite__update()'

    print('SUCCESSFUL TEST: antco.pheromone.updateDirEliteAS()')


def test_undirected_AS_elite_update():
    """ antco.pheromone.updateUndEliteAS() unit testing """
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T)/2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 1, 1],
         [1, 0, 0, 0],
         [1, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 2
        [[0, 1, 0, 1],
         [1, 0, 1, 0],
         [0, 1, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 1],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)
    expected = np.array([
        [0.0, 0.6414344987114436, 0.3820893643835099, 1.2433082310436099], 
        [0.6414344987114436, 0.0, 0.7473326730988266, 0.5720108649925117], 
        [0.3820893643835099, 0.7473326730988266, 0.0, 0.7692736491472838], 
        [1.2433082310436099, 0.5720108649925117, 0.7692736491472838, 0.0]],
        dtype=np.float64)

    ant_scores = np.array([0.2, 0.3, 0.4]).astype(np.float64)

    updateUndEliteAS(
        paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, elite=2, weight=1.0)
    
    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateUndEliteAS()'

    print('SUCCESSFUL TEST: antco.pheromone.updateUndEliteAS()')


def test_directed_MMAS_update():
    """ aco.pheromone.directed_mmas_update() unit testing """
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T) / 2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 2
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)
    expected = np.array([
        [0.0, 0.4267931249792329, 0.4776117072586296, 1.1791352931971335],
        [0.4267931249792329, 0.0, 0.5591658434565883, 0.3150135839042728],
        [0.4776117072586296, 0.5591658434565883, 0.0, 0.5865920636193305],
        [1.1791352931971335, 0.7150135839042728, 0.5865920636193305, 0.0]], np.float64)

    ant_scores = np.array([0.2, 0.3, 0.4], dtype=np.float64)

    updateDirMMAS(
        paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, limits=(0, 2), weight=1.0)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateDirMMAS()'

    expected2 = np.array([
        [0.0, 0.34143449871144366, 0.3820893643835099, 1.1433082310436098],
        [0.34143449871144366, 0.0, 0.4473326730988265, 0.2520108661846046],
        [0.3820893643835099, 0.4473326730988265, 0.0, 0.6692736491472838],
        [1.1433082310436098, 0.7720108649925117, 0.6692736491472838, 0.0]], np.float64)

    updateDirMMAS(
        paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, limits=(0, 2), weight=0.5)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected2, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateDirMMAS()'

    print('SUCCESSFUL TEST: antco.pheromone.updateDirMMAS()')


def test_directed_MMAS_elite_update():
    """ aco.pheromone.directed_mmas_elite_update() unit testing """
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T) / 2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 2
        [[0, 1, 1, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)

    expected = np.array([
        [0.0, 0.6414344987114436, 0.6820893643835099, 1.2433082310436099],
        [0.6414344987114436, 0.0, 0.4473326730988265, 0.2520108661846046],
        [0.3820893643835099, 0.4473326730988265, 0.0, 0.7692736491472838],
        [1.2433082310436099, 0.5720108649925117, 0.7692736491472838, 0.0]], np.float64)

    ant_scores = np.array([0.2, 0.3, 0.4], dtype=np.float64)

    updateDirEliteMMAS(
        paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, limits=(0, 2), elite=2,
        weight=1.0)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateDirEliteMMAS()'

    print('SUCCESSFUL TEST: antco.pheromone.updateDirEliteMMAS()')


def test_undirected_MMAS_update():
    """ aco.pheromone.undirected_mmas_update() unit testing """
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T) / 2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 2
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 1],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)

    expected = np.array([
        [0.0, 0.4267931249792329, 0.4776117072586296, 1.1791352931971335], 
        [0.4267931249792329, 0.0, 0.5591658434565883, 0.7150135839042728], 
        [0.4776117072586296, 0.5591658434565883, 0.0, 0.5865920636193305], 
        [1.1791352931971335, 0.7150135839042728, 0.5865920636193305, 0.0]], 
        dtype=np.float64)

    ant_scores = np.array([0.2, 0.3, 0.4], dtype=np.float64)

    updateUndMMAS(
        paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, limits=(0, 2), weight=1.0)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateUndMMAS()'

    expected2 = np.array([
        [0.0, 0.34143449871144366, 0.3820893643835099, 1.1433082310436098],
        [0.34143449871144366, 0.0, 0.4473326730988265, 0.7720108649925117],
        [0.3820893643835099, 0.4473326730988265, 0.0, 0.6692736491472838],
        [1.1433082310436098, 0.7720108649925117, 0.6692736491472838, 0.0]],
        dtype=np.float64)

    updateUndMMAS(
        paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, limits=(0, 2), weight=0.5)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected2, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateUndMMAS()'

    print('SUCCESSFUL TEST: antco.pheromone.undirected_mmas_update()')


def test_undirected_MMAS_elite_update():
    """ antco.pheromone.updateUndEliteMMAS() unit testing """
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T) / 2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 2
        [[0, 1, 0, 1],
         [1, 0, 1, 0],
         [0, 1, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 1],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)

    expected = np.array([
        [0.0, 0.6414344987114436, 0.3820893643835099, 1.2433082310436099], 
        [0.6414344987114436, 0.0, 0.7473326730988266, 0.5720108649925117], 
        [0.3820893643835099, 0.7473326730988266, 0.0, 0.7692736491472838], 
        [1.2433082310436099, 0.5720108649925117, 0.7692736491472838, 0.0]],
        dtype=np.float64)

    ant_scores = np.array([0.2, 0.3, 0.4], dtype=np.float64)
    
    updateUndEliteMMAS(
        paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, limits=(0, 2), elite=2, weight=1.0)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateUndEliteMMAS()'

    print('SUCCESSFUL TEST: antco.pheromone.updateUndEliteMMAS()')


def test_undirected_local_update():
    np.random.seed(1997)
    decay = 0.2
    init_val = 1.0
    P = np.random.uniform(low=1.0, high=3.0, size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P, 0)
    P = (P + P.T) / 2  # Symmetric matrix
    P_t0 = deepcopy(P)

    ant1 = Ant(l_min=0, l_max=5, graph_type='u'); ant1.initAdjMatrix(4)
    ant2 = Ant(l_min=0, l_max=5, graph_type='u'); ant2.initAdjMatrix(4)
    ant3 = Ant(l_min=0, l_max=5, graph_type='u'); ant3.initAdjMatrix(4)
    ant1.visited_nodes = [0, 2, 3]
    ant2.visited_nodes = [0, 1, 2]
    ant3.visited_nodes = [3, 0, 2]

    updateUndLocalPher(ant1, P, decay, init_val)
    assert P_t0[0, 0] == P[0, 0], 'FAILED TEST: antco.pheromone.updateUndLocalPher()'
    assert P_t0[2, 3] > P[2, 3], 'FAILED TEST: antco.pheromone.updateUndLocalPher()'
    assert P_t0[3, 2] > P[3, 2], 'FAILED TEST: antco.pheromone.updateUndLocalPher()'
    P_t0 = deepcopy(P)
    updateUndLocalPher(ant2, P, decay, init_val)
    assert P_t0[2, 3] == P[2, 3], 'FAILED TEST: antco.pheromone.updateUndLocalPher()'
    assert P_t0[1, 2] > P[1, 2], 'FAILED TEST: antco.pheromone.updateUndLocalPher()'
    assert P_t0[2, 1] > P[2, 1], 'FAILED TEST: antco.pheromone.updateUndLocalPher()'
    P_t0 = deepcopy(P)
    updateUndLocalPher(ant3, P, decay, init_val)
    assert P_t0[1, 2] == P[1, 2], 'FAILED TEST: antco.pheromone.updateUndLocalPher()'
    assert P_t0[0, 2] > P[0, 2], 'FAILED TEST: antco.pheromone.updateUndLocalPher()'
    assert P_t0[2, 0] > P[2, 0], 'FAILED TEST: antco.pheromone.updateUndLocalPher()'

    print('SUCCESSFUL TEST: antco.pheromone.updateUndLocalPher()')


def test_directed_local_update():
    np.random.seed(1997)
    decay = 0.2
    init_val = 1.0
    P = np.random.uniform(low=1.0, high=3.0, size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P, 0)
    P = (P + P.T) / 2  # Symmetric matrix
    P_t0 = deepcopy(P)

    ant1 = Ant(l_min=0, l_max=5, graph_type='d'); ant1.initAdjMatrix(4)
    ant2 = Ant(l_min=0, l_max=5, graph_type='u'); ant2.initAdjMatrix(4)
    ant3 = Ant(l_min=0, l_max=5, graph_type='u'); ant3.initAdjMatrix(4)
    ant1.visited_nodes = [0, 2, 3]
    ant2.visited_nodes = [0, 1, 2]
    ant3.visited_nodes = [3, 0, 2]

    updateDirLocalPher(ant1, P, decay, init_val)
    assert P_t0[0, 0] == P[0, 0], 'FAILED TEST: antco.pheromone.updateDirLocalPher()'
    assert P_t0[2, 3] > P[2, 3], 'FAILED TEST: antco.pheromone.updateDirLocalPher()'
    assert P_t0[3, 2] == P[3, 2], 'FAILED TEST: antco.pheromone.updateDirLocalPher()'
    P_t0 = deepcopy(P)
    updateDirLocalPher(ant2, P, decay, init_val)
    assert P_t0[2, 3] == P[2, 3], 'FAILED TEST: antco.pheromone.updateDirLocalPher()'
    assert P_t0[1, 2] > P[1, 2], 'FAILED TEST: antco.pheromone.updateDirLocalPher()'
    assert P_t0[2, 1] == P[2, 1], 'FAILED TEST: antco.pheromone.updateDirLocalPher()'
    P_t0 = deepcopy(P)
    updateDirLocalPher(ant3, P, decay, init_val)
    assert P_t0[1, 2] == P[1, 2], 'FAILED TEST: antco.pheromone.updateDirLocalPher()'
    assert P_t0[0, 2] > P[0, 2], 'FAILED TEST: antco.pheromone.updateDirLocalPher()'
    assert P_t0[2, 0] == P[2, 0], 'FAILED TEST: antco.pheromone.updateDirLocalPher()'

    print('SUCCESSFUL TEST: antco.pheromone.updateDirLocalPher()')


def test_undirected_ACS():
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T) / 2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 0, 1],
         [1, 0, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 1, 0]],
        # Ant 2
        [[0, 1, 0, 1],
         [1, 0, 1, 0],
         [0, 1, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 1],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)

    expected = np.array([
        [0.0, 0.5334914082114515, 0.5970146362973399, 1.1591352988595747],
        [0.5334914082114515, 0.0, 0.6989573069245543, 0.695013589566714],
        [0.5970146362973399, 0.6989573069245543, 0.0, 0.5665920692817716],
        [1.1591352988595747, 0.695013589566714, 0.5665920692817716, 0.0]],
        dtype=np.float64)

    ant_scores = np.array([0.2, 0.3, 1.9], dtype=np.float64)

    updateUndACS(paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, weight=1.0)
    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateUndACS()'

    print('SUCCESSFUL TEST: antco.pheromone.updateUndACS()')


def test_directed_ACS():
    np.random.seed(1997)
    evaporation = 0.2
    P_t0 = np.random.uniform(size=(4, 4)).astype(np.float64)
    np.fill_diagonal(P_t0, 0)
    P_t0 = (P_t0 + P_t0.T) / 2  # Symmetric matrix
    paths = np.array([
        # Ant 1
        [[0, 1, 0, 1],
         [0, 0, 0, 0],
         [0, 0, 0, 1],
         [0, 0, 0, 0]],
        # Ant 2
        [[0, 1, 0, 1],
         [1, 0, 1, 0],
         [0, 1, 0, 1],
         [1, 0, 1, 0]],
        # Ant 3
        [[0, 0, 0, 1],
         [0, 0, 0, 1],
         [0, 0, 0, 1],
         [1, 1, 1, 0]]], dtype=np.int8)

    expected = np.array([
        [0.0, 0.6667931285555115, 0.5970146362973399, 1.0191352967734122],
        [0.5334914082114515, 0.0, 0.6989573069245543, 0.3937669813472373],
        [0.5970146362973399, 0.6989573069245543, 0.0, 0.42659206719560916],
        [0.9739191201245483, 0.3937669813472373, 0.23324008039305005, 0.0]],
        dtype=np.float64)

    ant_scores = np.array([0.8, 0.3, 0.2], dtype=np.float64)

    updateDirACS(paths=paths, P=P_t0, ant_scores=ant_scores, rho=evaporation, weight=1.5)

    assert np.all(np.round(P_t0, decimals=4) == np.round(expected, decimals=4)), \
        'FAILED TEST: antco.pheromone.updateDirACS()'

    print('SUCCESSFUL TEST: antco.pheromone.updateDirACS()')


def test():
    test_directed_AS_update()
    test_undirected_AS_update()
    test_directed_AS_elite_update()
    test_undirected_AS_elite_update()
    test_directed_MMAS_update()
    test_directed_MMAS_elite_update()
    test_undirected_MMAS_update()
    test_undirected_MMAS_elite_update()
    test_undirected_local_update()
    test_directed_local_update()
    test_undirected_ACS()
    test_directed_ACS()



# Module that collects the possible preprocessing steps applicable to an antco.ACO instance prior
# to performing the optimisation.
#
# Author: Fernando García Gutiérrez
# Email: fegarc05@ucm.es
#
import numpy as np
from .aco import ACO
from .c_utils import minMaxScaling


def apply(aco_obj: ACO, **kwargs):
    """
    Function that applies the specified transformations (passed as a string) to the ACO instance.

    Parameters
    ----------
    aco_obj: antco.ACO
        Ant Colony Optimization (aka ACO) instance on which the transformations are to be applied.

    **kwargs
        Transformations to be applied to the ACO instance. Only the indicated transformations will
        be applied.

        scale_heuristic: dict
            Scales the heuristic matrix to a range of values defined by [min_val, max_val]

            >>> apply(aco_obj, scale_heuristic={'min_val': 1.0, 'max_val': 2.0})

        accessory_node: bool
            Indicates whether to add an accessory node densely connected to all nodes in the
            network. This way all the ants will start from this accessory node, optimising the
            initial positioning in the network.

            >>> apply(aco_obj, accessory_node=True)

    """
    if 'scale_heuristic' in kwargs:
        _scaleHeuristic(aco_obj, **kwargs['scale_heuristic'])

    if 'accessory_node' in kwargs and kwargs['accessory_node'] is True:
        _addAccessoryNode(aco_obj)


def _scaleHeuristic(aco_obj: ACO, min_val: float = 0.0, max_val: float = 1.0, **_):
    """
    Function that scales the heuristic matrix to a specific range defined by min_val and max_val.

    Parameters
    ----------
    aco_obj: ACO:
        ACO instance.

    min_val: float, default=0.0
        Min value.

    max_val: float, default=1.0
        Maximum value.
    """
    heuristic = aco_obj.heuristic
    heuristic = minMaxScaling(heuristic, aco_obj.graph)
    heuristic[aco_obj.graph == 1] = heuristic[aco_obj.graph == 1] * (max_val - min_val) + min_val
    aco_obj.heuristic = (heuristic, True)


def _addAccessoryNode(aco_obj: ACO):
    """ Method that adds an accessory node to the heuristics, pheromone and connectivity graph
    matrices. """
    def _add_node(matrix: np.ndarray):
        """ Adds an extra densely connected node to all nodes of the matrix in the last position. """
        new_shape = matrix.shape[0] + 1, matrix.shape[1] + 1
        new_matrix = np.empty(new_shape, dtype=matrix.dtype)

        # Copy matrix values
        for i in range(matrix.shape[0]):
            for j in range(matrix.shape[1]):
                new_matrix[i, j] = matrix[i, j]

        # Add node using the maximum value
        new_matrix[:, -1] = np.max(matrix)
        new_matrix[-1, :] = np.max(matrix)

        return new_matrix

    graph = _add_node(aco_obj.graph)
    heuristic = _add_node(aco_obj.heuristic)
    pheromones = _add_node(aco_obj.pheromones)

    assert graph.shape[0] == graph.shape[1] == heuristic.shape[0] == heuristic.shape[1] == pheromones.shape[0] == \
           pheromones.shape[1], 'Incorrect shapes adding accessory node.'

    # Accessory node will be in the last position
    aco_obj.accessory_node = graph.shape[0] - 1
    aco_obj.graph = (graph, True)
    aco_obj.heuristic = (heuristic, True)
    aco_obj.pheromones = (pheromones, True)
    aco_obj.objectiveFunction.accessory_node = True
